# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import hdtv.cal
import hdtv.color
import hdtv.ui
from hdtv.util import LockViewport


class Drawable(object):
    def __init__(self, color=None, cal=None):
        self.viewport = None
        # displayObj will be created when calling Draw
        self.displayObj = None
        self.active = False
        self.cal = cal
        self.color = color
        self.ID = None

    @property
    def name(self):
        return str(self.displayObj)

    # cal property
    def _set_cal(self, cal):
        self._cal = hdtv.cal.MakeCalibration(cal)
        # update display if needed
        try:
            self.displayObj.SetCal(self._cal)
        except BaseException:
            pass

    def _get_cal(self):
        return self._cal

    cal = property(_get_cal, _set_cal)

    # color property
    def _set_color(self, color):
        self._activeColor = hdtv.color.Highlight(color, active=True)
        self._passiveColor = hdtv.color.Highlight(color, active=False)
        # update display if needed
        if self.displayObj is not None:
            if self._active:
                self.displayObj.SetColor(self._activeColor)
            else:
                self.displayObj.SetColor(self._passiveColor)

    def _get_color(self):
        return self._passiveColor

    color = property(_get_color, _set_color)

    # active property
    def _set_active(self, state):
        self._active = state
        if self.displayObj is not None:
            if self._active:
                self.displayObj.SetColor(self._activeColor)
                # move the object to the top of its draw stack
                try:
                    self.displayObj.ToTop()
                except BaseException:
                    pass
            else:
                self.displayObj.SetColor(self._passiveColor)

    def _get_active(self):
        return self._active

    active = property(_get_active, _set_active)

    # ID property
    def _set_ID(self, ID):
        self._ID = ID
        if self.displayObj:
            try:
                ID = str(ID).strip(".")
                self.displayObj.SetID(ID)
            except BaseException:
                pass

    def _get_ID(self):
        return self._ID

    ID = property(_get_ID, _set_ID)

    def Draw(self, viewport):
        """
        This function must create the appropriate object from the underlying
        C++ display code and call its draw function.
        Attention: Unlike the Display object of the underlying implementation,
        python objects can only be drawn on a single viewport
        """
        self.viewport = viewport

    def Refresh(self):
        """
        Refresh the objects data
        """
        pass

    def Show(self):
        """
        Show the object
        """
        if not self.viewport:
            return
        if self.displayObj:
            if self.active:
                self.displayObj.SetColor(self._activeColor)
                # move the object to the top of its draw stack
                try:
                    self.displayObj.ToTop()
                except BaseException:
                    pass
            else:
                self.displayObj.SetColor(self._passiveColor)
            if self.cal:
                self.displayObj.SetCal(self.cal)
            self.displayObj.Show()

    def Hide(self):
        """
        Hide the object
        """
        if not self.viewport:
            return
        if self.displayObj:
            self.displayObj.Hide()


class DrawableManager(object):
    """
    This class provides some handy functions to manage a collection of
    identical drawable objects.
    """

    def __init__(self, viewport=None):
        self.viewport = viewport
        # dictionary to store the drawable objects
        self.dict = dict()
        self.visible = set()
        self.activeID = None
        # This should keep track of ID for nextID, prevID
        self._iteratorID = self.activeID
        self._active = False

    def __len__(self):
        return len(self.dict)

    @property
    def ids(self):
        # return sorted list of ids
        ids = sorted(self.dict.keys())
        return ids

    # active property
    def _set_active(self, state):
        self._active = state
        if self.activeID is not None:
            # give state to the active child
            self.GetActiveObject().active = state

    def _get_active(self):
        return self._active

    active = property(_get_active, _set_active)

    def ActivateObject(self, ID=None):
        """
        Activates the object with ID
        """
        if ID is not None and ID not in self.ids:
            raise KeyError
        with LockViewport(self.viewport):
            # change state of former active object
            if self.activeID is not None:
                self.GetActiveObject().active = False
            # activate new object
            self.activeID = ID
            if self.activeID is not None:
                # reset iterator
                self._iteratorID = self.activeID
                # change state of object
                self.GetActiveObject().active = self.active
                # call ShowObject, to make sure the new active object is visible
                self.ShowObjects(ID, clear=False)

    def GetActiveObject(self):
        """
        Returns currently active object
        """
        if self.activeID is None:
            return None
        else:
            return self.dict[self.activeID]

    def Get(self, ID):
        #        if not isinstance(ID, str):
        #            ID = str(ID)
        ID = hdtv.util.ID.ParseIds(ID, self)[0]
        return self.dict[ID]

    def Index(self, obj):
        """
        Return index such that self[index] == obj
        """
        index = [k for (k, v) in self.dict.items() if v == obj]
        if len(index) == 0:
            raise ValueError("Object not found in this collection")
        else:
            return index[0]

    def Insert(self, obj, ID=None):
        """
        This inserts an object to the dictionary of this manager
        If no ID is given, the first free ID is used, else the object is inserted
        at the given ID, possibly removing an object which was there before.
        """
        # if no ID is specified we take the first free ID
        if ID is None:
            ID = self.GetFreeID()
        self._iteratorID = ID
        self.dict[ID] = obj
        obj.ID = ID
        if self.viewport:
            obj.Draw(self.viewport)
            self.visible.add(ID)
        return ID

    def Pop(self, ID):
        """
        Remove object with ID
        """
        if ID == self.activeID:
            self.ActivateObject(None)
        if ID == self._iteratorID:
            # set iterator to the ID before the one we remove
            self._iteratorID = self.prevID
        self.visible.discard(ID)
        try:
            obj = self.dict.pop(ID)
            obj.ID = None
            return obj
        except KeyError:
            hdtv.ui.warning("ID %s not found." % ID)

    def Clear(self):
        """
        Clear dict and reset everything
        """
        self.activeID = None
        self._iterator = self.activeID
        self.visible.clear()
        self.dict.clear()

    def GetFreeID(self):
        """
        Finds the first free index
        """
        ids = [i.major for i in self.ids]
        ID = 0
        while ID in ids:
            ID += 1
        return hdtv.util.ID(major=ID)

    def Draw(self, viewport):
        """
        Draw function (sets the viewport and draws all components)
        """
        if not self.viewport is None and not self.viewport == viewport:
            # Unlike the Display object of the underlying implementation,
            # python objects can only be drawn on a single viewport
            raise RuntimeError("Object can only be drawn on a single viewport")
        self.viewport = viewport
        with LockViewport(self.viewport):
            for ID in self.dict.keys():
                self.dict[ID].Draw(self.viewport)
                self.visible.add(ID)

    # Refresh commands
    def Refresh(self):
        """
        Refresh whole object
        """
        return self.RefreshAll()

    def RefreshAll(self):
        """
        Refresh all objects in dict
        """
        return self.RefreshObjects(iter(self.dict.keys()))

    def RefreshVisible(self):
        """
        Refresh visible objects
        """
        return self.RefreshObjects(self.visible)

    def RefreshObjects(self, ids):
        """
        Refresh objects with ids
        """
        with LockViewport(self.viewport):
            try:
                iter(ids)
            except BaseException:
                ids = [ids]
            for ID in ids:
                try:
                    self.dict[ID].Refresh()
                except KeyError:
                    hdtv.ui.warning("ID %d not found" % ID)
            return ids

    # Hide commands
    def Hide(self):
        """
        Hide the object as a whole (with info about current visible/active states)
        """
        # do not call HideAll here, as then we loose the info about
        # visible/active states of the objects
        for obj in self.dict.values():
            obj.Hide()

    def HideAll(self):
        """
        Hide all child objects
        """
        return self.HideObjects(list(self.dict.keys()))

    def HideObjects(self, ids):
        """
        Hide objects
        """
        if self.viewport is None:
            return
        with LockViewport(self.viewport):
            # check if just single id
            try:
                iter(ids)
            except BaseException:
                ids = [ids]
            for ID in ids:
                try:
                    self.dict[ID].Hide()
                    self.visible.discard(ID)
                except KeyError:
                    hdtv.ui.warning("ID %d not found" % ID)
            return ids

    # Show commands:
    def Show(self):
        """
        Show the object as whole (according to last visible/active states)
        """
        self.ShowObjects(self.visible)
        if self.activeID is not None:
            self.GetActiveObject().active = self.active

    def ShowAll(self):
        """
        Show all
        """
        return self.ShowObjects(list(self.dict.keys()))

    def ShowObjects(self, ids, clear=True):
        """
        Show objects on the display.

        If the clear parameter is True, the display is cleared first.
        Otherwise the objects are shown in addition to the ones, that
        are already visible.
        """
        if self.viewport is None:
            return
        with LockViewport(self.viewport):
            # check if just single id
            try:
                iter(ids)
            except BaseException:
                ids = [ids]
            if clear:
                # hide all other objects except in ids
                # do not use HideAll, because if the active objects is among
                # the objects that should be shown, its state would be lost
                others = set(self.dict.keys()) - set(ids)
                self.HideObjects(others)
            for ID in ids:
                try:
                    self.dict[ID].Show()
                    self.visible.add(ID)
                except KeyError:
                    hdtv.ui.warning("ID %s not found" % ID)
            return ids

    # nextID/prevID/firstID/lastID getter
    @property
    def nextID(self):
        return self._nextID(onlyVisible=False)

    @property
    def nextVisibleID(self):
        return self._nextID(onlyVisible=True)

    @property
    def prevID(self):
        return self._prevID(onlyVisible=False)

    @property
    def prevVisibleID(self):
        return self._prevID(onlyVisible=True)

    @property
    def firstID(self):
        return self._firstID(onlyVisible=False)

    @property
    def firstVisibleID(self):
        return self._firstID(onlyVisible=True)

    @property
    def lastID(self):
        return self._lastID(onlyVisible=False)

    @property
    def lastVisibleID(self):
        return self._lastID(onlyVisible=True)

    def _firstID(self, onlyVisible=False):
        if onlyVisible:
            ids = list(self.visible)
        else:
            ids = list(self.dict.keys())
        ids.sort()

        try:
            firstID = min(ids)
        except ValueError:
            firstID = self.activeID

        self._iteratorID = firstID
        hdtv.ui.debug("hdtv.drawable.DrawableManager: firstID=" + str(firstID), level=6)
        return firstID

    def _lastID(self, onlyVisible=False):
        if onlyVisible:
            ids = list(self.visible)
        else:
            ids = list(self.dict.keys())
        ids.sort()

        try:
            lastID = max(ids)
        except ValueError:
            lastID = self.activeID

        self._iteratorID = lastID
        hdtv.ui.debug("hdtv.drawable.DrawableManager: lastID=" + str(lastID), level=6)
        return lastID

    def _nextID(self, onlyVisible=False):
        """
        Get next ID after _iteratorID
        """
        try:
            if onlyVisible:
                ids = list(self.visible)
            else:
                ids = list(self.dict.keys())

            ids.sort()
            nextIndex = (ids.index(self._iteratorID) + 1) % len(ids)
            nextID = ids[nextIndex]

        except ValueError:
            nextID = self.activeID if self.activeID is not None else self.firstID

        hdtv.ui.debug("hdtv.drawable.DrawableManager: nextID=" + str(nextID), level=6)
        self._iteratorID = nextID
        return nextID

    def _prevID(self, onlyVisible=False):
        """
        Get previous ID before _iteratorID
        """
        try:
            if onlyVisible:
                ids = list(self.visible)
            else:
                ids = list(self.dict.keys())

            ids.sort()
            prevIndex = (ids.index(self._iteratorID) - 1) % len(ids)
            prevID = ids[prevIndex]
        except ValueError:
            prevID = self.activeID if self.activeID is not None else self.lastID

        self._iteratorID = prevID
        hdtv.ui.debug("hdtv.drawable.DrawableManager: prevID=" + str(prevID), level=6)
        return prevID

    def ShowNext(self, nb=1):
        """
        Show next object (by index)

        With the parameter nb, the user can choose how many new spectra
        there should be displayed.
        """
        if nb > len(self.dict):
            self.ShowAll()
            return
        ids = self.ids
        index = ids.index(self.nextID)
        ids = ids[index : index + nb]
        self.ShowObjects(ids, clear=True)
        return ids

    def ShowPrev(self, nb=1):
        """
        Show previous object (by index)

        With the parameter nb, the user can choose how many new spectra
        there should be displayed.
        """
        if nb > len(self.dict):
            self.ShowAll()
            return
        ids = self.ids
        index = ids.index(self.prevID)
        ids = ids[index : index + nb]
        self.ShowObjects(ids, clear=True)
        return ids

    def ShowFirst(self, nb=1):
        """
        Show the first nb objects
        """
        if nb > len(self.dict):
            return self.ShowAll()
        ids = self.ids
        ids = ids[:nb]
        self.ShowObjects(ids, clear=True)
        return ids

    def ShowLast(self, nb=1):
        """
        Show the last nb objects
        """
        if nb > len(self.dict):
            self.ShowAll()
            return
        ids = self.ids
        ids = ids[len(ids) - nb :]
        self.ShowObjects(ids, clear=True)
        return ids



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import math
from functools import total_ordering
from uncertainties import ufloat

import ROOT
import hdtv.rootext.display
import hdtv.rootext.fit

from .peak import PeakModel
from hdtv.drawable import Drawable
import hdtv.options


@total_ordering
class TheuerkaufPeak(Drawable):
    """
    Peak object for the Theuerkauf (classic TV) fitter
    """

    def __init__(self, pos, vol, width, tl, tr, sh, sw, color=None, cal=None):
        super(TheuerkaufPeak, self).__init__(color, cal)
        # values are uncalibrated!
        self.pos = pos
        self.vol = vol
        # width==fwhm (internally the C++ fitter uses sigma)
        self.width = width
        self.tl = tl
        self.tr = tr
        self.sh = sh
        self.sw = sw
        # dictionary for storing additional user supplied values
        self.extras = dict()

    def __getattr__(self, name):
        """
        calculate calibrated values on the fly for pos and width
        """
        if name == "pos_cal":
            if self.cal is None:
                return self.pos
            pos_uncal = self.pos.nominal_value
            pos_err_uncal = self.pos.std_dev
            pos_cal = self.cal.Ch2E(pos_uncal)
            pos_err_cal = abs(self.cal.dEdCh(pos_uncal) * pos_err_uncal)
            return ufloat(pos_cal, pos_err_cal, self.pos.tag)
        elif name == "width_cal":
            if self.cal is None:
                return self.width
            pos_uncal = self.pos.nominal_value
            hwhm_uncal = self.width.nominal_value / 2
            width_err_uncal = self.width.std_dev
            width_cal = self.cal.Ch2E(pos_uncal + hwhm_uncal) - self.cal.Ch2E(
                pos_uncal - hwhm_uncal
            )
            # This is only an approximation, valid as d(width_cal)/d(pos_uncal) \approx 0
            #  (which is true for Ch2E \approx linear)
            width_err_cal = abs(
                (
                    self.cal.dEdCh(pos_uncal + hwhm_uncal) / 2.0
                    + self.cal.dEdCh(pos_uncal - hwhm_uncal) / 2.0
                )
                * width_err_uncal
            )
            return ufloat(width_cal, width_err_cal, self.width.tag)
        elif name in ["vol_cal", "tl_cal", "tr_cal", "sh_cal", "sw_cal"]:
            name = name[0 : name.rfind("_cal")]
            return getattr(self, name)
        else:
            # DON'T FORGET THIS LINE! see
            # http://code.activestate.com/recipes/52238/
            raise AttributeError(name)

    def __str__(self):
        return self.formatted_str(verbose=False)

    def formatted_str(self, verbose=False):
        """
        print the properties of this peak in a nicely formatted way
        """
        text = str()
        if verbose:
            text += "Pos:         {0.pos_cal:S}\n"
            text += "Channel:     {0.pos:S}\n"
            text += "Volume:      {0.vol:S}\n"
            text += "FWHM:        {0.width_cal:S}\n"
            if self.tl is not None:
                text += "Left Tail:   {0.tl:S}\n"
            else:
                text += "Left Tail:   None\n"
            if self.tr is not None:
                text += "Right Tail:  {0.tr:S}\n"
            else:
                text += "Right Tail:  None\n"
            if self.sh is not None:
                text += "Step height: {0.sh:S}\n"
                text += "Step width:  {0.sw:S}\n"
            else:
                text += "Step:        None"
        else:
            text += "Peak@ {0.pos_cal}"
        return text.format(self)

    def __eq__(self, other):
        return (self.pos, self.vol, self.width, self.tl, self.tr, self.sh, self.sw) == (
            other.pos,
            other.vol,
            other.width,
            other.tl,
            other.tr,
            other.sh,
            other.sw,
        )

    def __ne__(self, other):
        return not (
            (self.pos, self.vol, self.width, self.tl, self.tr, self.sh, self.sw)
            == (
                other.pos,
                other.vol,
                other.width,
                other.tl,
                other.tr,
                other.sh,
                other.sw,
            )
        )

    def __lt__(self, other):
        return self.pos.nominal_value < other.pos.nominal_value

    def Draw(self, viewport):
        """
        Draw the function of this peak
        """
        if self.viewport:
            if self.viewport == viewport:
                self.Show()
            else:
                # Unlike the Display object of the underlying implementation,
                # python objects can only be drawn on a single viewport
                raise RuntimeError("Peak cannot be drawn on multiple viewports")
        self.viewport = viewport
        if self.displayObj:
            self.displayObj.Draw(self.viewport)


class PeakModelTheuerkauf(PeakModel):
    """
    Theuerkauf peak model - "classical" model used by tv
    """

    def __init__(self):
        super(PeakModelTheuerkauf, self).__init__()
        self.fParStatus = {
            "pos": None,
            "vol": None,
            "width": None,
            "tl": None,
            "tr": None,
            "sh": None,
            "sw": None,
        }
        self.fValidParStatus = {
            "pos": [float, "free", "hold"],
            "vol": [float, "free", "hold"],
            "width": [float, "free", "equal"],
            "tl": [float, "free", "equal", "none"],
            "tr": [float, "free", "equal", "none"],
            "sh": [float, "free", "equal", "none"],
            "sw": [float, "free", "equal", "hold"],
        }
        self.fOptStatus = {
            "integrate": False,
            "likelihood": "normal",
        }
        self.fValidOptStatus = {
            "integrate": [False, True],
            "likelihood": ["normal", "poisson"],
        }

        self.ResetParamStatus()
        self.Peak = TheuerkaufPeak
        self.name = "theuerkauf"

    def CopyPeak(self, cpeak, color=None, cal=None):
        """
        create a python peak object from C++ peak object
        """
        # get values from C++ object (uncalibrated)
        pos_uncal = cpeak.GetPos()
        pos_err_uncal = cpeak.GetPosError()
        # width==fwhm (internally the C++ fitter uses sigma)
        width_uncal = cpeak.GetSigma() * 2.0 * math.sqrt(2.0 * math.log(2.0))
        width_err_uncal = cpeak.GetSigmaError() * 2.0 * math.sqrt(2.0 * math.log(2.0))
        # create ufloats from this
        pos = ufloat(pos_uncal, pos_err_uncal, cpeak.PosIsFree())
        vol = ufloat(cpeak.GetVol(), cpeak.GetVolError(), cpeak.VolIsFree())
        width = ufloat(width_uncal, width_err_uncal, cpeak.SigmaIsFree())
        # optional parameters
        if cpeak.HasLeftTail():
            tl = ufloat(
                cpeak.GetLeftTail(), cpeak.GetLeftTailError(), cpeak.LeftTailIsFree()
            )
        else:
            tl = None
        if cpeak.HasRightTail():
            tr = ufloat(
                cpeak.GetRightTail(), cpeak.GetRightTailError(), cpeak.RightTailIsFree()
            )
        else:
            tr = None
        if cpeak.HasStep():
            sh = ufloat(
                cpeak.GetStepHeight(),
                cpeak.GetStepHeightError(),
                cpeak.StepHeightIsFree(),
            )
            sw = ufloat(
                cpeak.GetStepWidth(), cpeak.GetStepWidthError(), cpeak.StepWidthIsFree()
            )
        else:
            sh = sw = None
        # create peak object
        peak = self.Peak(pos, vol, width, tl, tr, sh, sw, color, cal)
        func = cpeak.GetPeakFunc()
        peak.displayObj = ROOT.HDTV.Display.DisplayFunc(func, color)
        peak.displayObj.SetCal(cal)
        return peak

    def RestoreParams(self, peak, cpeak):
        """
        Restore the params of a C++ peak object using a python peak object
        """
        cpeak.RestorePos(peak.pos.nominal_value, peak.pos.std_dev)
        cpeak.RestoreVol(peak.vol.nominal_value, peak.vol.std_dev)
        # internally the C++ fitter uses sigma
        sigma = peak.width.nominal_value / (2.0 * math.sqrt(2.0 * math.log(2.0)))
        sigma_err = peak.width.std_dev / (2.0 * math.sqrt(2.0 * math.log(2.0)))
        cpeak.RestoreSigma(sigma, sigma_err)
        if peak.tl:
            cpeak.RestoreLeftTail(peak.tl.nominal_value, peak.tl.std_dev)
        if peak.tr:
            cpeak.RestoreRightTail(peak.tr.nominal_value, peak.tr.std_dev)
        if peak.sh:
            cpeak.RestoreStepHeight(peak.sh.nominal_value, peak.sh.std_dev)
            cpeak.RestoreStepWidth(peak.sw.nominal_value, peak.sw.std_dev)

    def ResetParamStatus(self):
        """
        Reset parameter status to defaults
        """
        self.fParStatus["pos"] = "free"
        self.fParStatus["vol"] = "free"
        self.fParStatus["width"] = "equal"
        self.fParStatus["tl"] = "none"
        self.fParStatus["tr"] = "none"
        self.fParStatus["sh"] = "none"
        self.fParStatus["sw"] = "hold"
        self.fOptStatus["integrate"] = False
        self.fOptStatus["likelihood"] = "normal"

    def Uncal(self, parname, value, pos_uncal, cal):
        """
        Convert a value from calibrated (spectrum) to uncalibrated (fitter) units
        This is needed, when a value is hold to a specific calibrated value.
        """
        if parname == "pos":
            return cal.E2Ch(value)
        elif parname == "width":
            pos_cal = cal.Ch2E(pos_uncal)
            width_uncal = cal.E2Ch(pos_cal + value / 2.0) - cal.E2Ch(
                pos_cal - value / 2.0
            )
            # Note that the underlying fitter uses ``sigma'' as a parameter
            #  (see HDTV technical documentation in the wiki)
            return width_uncal / (2.0 * math.sqrt(2.0 * math.log(2.0)))
        elif parname in ("vol", "tl", "tr", "sh", "sw"):
            return value
        else:
            raise RuntimeError("Unexpected parameter name")

    def GetFitter(self, region, peaklist, cal):
        """
        Creates a C++ Fitter object, which can then do the real work
        """
        # Define a fitter and a region
        # FIXME: show_inipar seems to create a crash, see ticket #103 for trace
        # debug_show_inipar = hdtv.options.Get("__debug__.fit.show_inipar")
        # self.fFitter = ROOT.HDTV.Fit.TheuerkaufFitter(region[0],region[1],debug_show_inipar)
        integrate = self.GetOption("integrate")
        likelihood = self.GetOption("likelihood")
        self.fFitter = ROOT.HDTV.Fit.TheuerkaufFitter(
            region[0], region[1], integrate, likelihood
        )
        self.ResetGlobalParams()
        # Check if enough values are provided in case of per-peak parameters
        #  (the function raises a RuntimeError if the check fails)
        self.CheckParStatusLen(len(peaklist))

        # Copy peaks to the fitter
        for pid in range(0, len(peaklist)):
            pos_uncal = peaklist[pid]

            pos = self.GetParam("pos", pid, pos_uncal, cal, pos_uncal)
            vol = self.GetParam("vol", pid, pos_uncal, cal)
            sigma = self.GetParam("width", pid, pos_uncal, cal)
            tl = self.GetParam("tl", pid, pos_uncal, cal)
            tr = self.GetParam("tr", pid, pos_uncal, cal)
            sh = self.GetParam("sh", pid, pos_uncal, cal)
            sw = self.GetParam("sw", pid, pos_uncal, cal)

            cpeak = ROOT.HDTV.Fit.TheuerkaufPeak(pos, vol, sigma, tl, tr, sh, sw)
            self.fFitter.AddPeak(cpeak)

        return self.fFitter



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import re
import os
import sys
import warnings
import shutil
from math import ceil

import pytest

from tests.helpers.utils import redirect_stdout, hdtvcmd
from tests.helpers.fixtures import temp_file

from hdtv.util import monkey_patch_ui

monkey_patch_ui()

import hdtv.cmdline
import hdtv.options
import hdtv.session

import __main__

try:
    __main__.spectra = hdtv.session.Session()
except RuntimeError:
    pass


import hdtv.plugins.specInterface

s = hdtv.plugins.specInterface.spec_interface
spectra = __main__.spectra

testspectrum = os.path.join(os.path.curdir, "tests", "share", "osiris_bg.spc")


@pytest.fixture(autouse=True)
def prepare():
    hdtv.options.Set("table", "classic")
    hdtv.options.Set("uncertainties", "short")
    spectra.Clear()
    yield
    spectra.Clear()


def test_cmd_spectrum_get():
    assert len(s.spectra.dict) == 0
    f, ferr = hdtvcmd("spectrum get {}".format(testspectrum))
    res_specfile, specid = re.search(r"Loaded (.*) into (\d+)", f).groups()
    assert len(s.spectra.dict) == 1
    assert res_specfile == testspectrum
    for (ID, obj) in s.spectra.dict.items():
        assert str(ID) == specid
        assert type(obj) == hdtv.spectrum.Spectrum


@pytest.mark.parametrize("slot", [3, 1000000])
def test_cmd_spectrum_get_slot(slot):
    assert len(s.spectra.dict) == 0
    f, ferr = hdtvcmd("spectrum get -s {} {}".format(slot, testspectrum))
    res_specfile, specid = re.search(r"Loaded (.*) into (\d+)", f).groups()
    assert len(s.spectra.dict) == 1
    assert res_specfile == testspectrum
    assert specid == str(slot)
    for (ID, obj) in s.spectra.dict.items():
        assert str(ID) == specid
        assert type(obj) == hdtv.spectrum.Spectrum


@pytest.mark.parametrize("num", [4, 40])
def test_cmd_spectrum_get_repeated(num):
    assert len(s.spectra.dict) == 0
    query = " ".join([testspectrum] * num)
    hdtvcmd("spectrum get {}".format(query))
    assert len(s.spectra.dict) == num


@pytest.mark.parametrize("pattern", ["tests/share/osiris_[a-z][a-z].spc"])
def test_cmd_spectrum_get_pattern(pattern):
    assert len(s.spectra.dict) == 0
    hdtvcmd("spectrum get {}".format(pattern))
    assert len(s.spectra.dict) == 1


# Loading cal as spec file. This is really stupid, but it works.
@pytest.mark.parametrize(
    "specfiles", [["tests/share/osiris_bg.spc", "tests/share/osiris_bg.cal"]]
)
def test_cmd_spectrum_get_multi(specfiles):
    assert len(s.spectra.dict) == 0
    query = " ".join(specfiles)
    hdtvcmd("spectrum get {}".format(query))
    assert len(s.spectra.dict) == len(specfiles)


@pytest.mark.parametrize("numspecs", [1, 10])
def test_cmd_spectrum_list(numspecs):
    assert len(s.spectra.dict) == 0
    for _ in range(numspecs):
        hdtvcmd("spectrum get {}".format(testspectrum))
    f, ferr = hdtvcmd("spectrum list")
    assert len(s.spectra.dict) == numspecs
    assert len(f.split("\n")) == numspecs + 2


def test_cmd_spectrum_delete():
    assert len(s.spectra.dict) == 0
    f, ferr = hdtvcmd("spectrum get {}".format(testspectrum))
    res_specfile, specid = re.search(r"Loaded (.*) into (\d+)", f).groups()

    assert len(s.spectra.dict) == 1

    f, ferr = hdtvcmd("spectrum delete {}".format(specid))
    assert len(s.spectra.dict) == 0


def test_cmd_spectrum_delete_none():
    assert len(s.spectra.dict) == 0
    hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == 1

    f, ferr = hdtvcmd("spectrum delete")
    assert "WARNING: Nothing to do" in ferr
    assert len(s.spectra.dict) == 1


@pytest.mark.parametrize("numspecs", [1, 5])
def test_cmd_spectrum_delete_all(numspecs):
    assert len(s.spectra.dict) == 0
    for _ in range(numspecs):
        hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == numspecs

    hdtvcmd("spectrum delete all")
    assert len(s.spectra.dict) == 0


@pytest.mark.parametrize("numspecs", [1, 5])
def test_cmd_spectrum_delete_several(numspecs):
    assert len(s.spectra.dict) == 0
    for _ in range(numspecs):
        hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == numspecs

    specs = " ".join(str(e) for e in range(numspecs))
    hdtvcmd("spectrum delete {}".format(specs))
    assert len(s.spectra.dict) == 0


@pytest.mark.parametrize("numspecs", [2, 5])
def test_cmd_spectrum_hide(numspecs):
    numhidespecs = max(1, numspecs // 2)
    assert len(s.spectra.dict) == 0
    for _ in range(numspecs):
        hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == numspecs

    specs = " ".join(str(e) for e in range(numhidespecs))
    hdtvcmd("spectrum hide {}".format(specs))

    f, ferr = hdtvcmd("spectrum list -v")
    assert len(s.spectra.dict) == numspecs
    assert len(f.split("\n")) == numspecs - numhidespecs + 2


@pytest.mark.parametrize("numspecs", [2, 5])
def test_cmd_spectrum_show(numspecs):
    numshowspecs = max(1, numspecs // 2)
    assert len(s.spectra.dict) == 0
    for _ in range(numspecs):
        hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == numspecs

    specs = " ".join(str(e) for e in range(numshowspecs))
    hdtvcmd("spectrum show {}".format(specs))

    f, ferr = hdtvcmd("spectrum list -v")
    assert len(s.spectra.dict) == numspecs
    assert len(f.split("\n")) == numshowspecs + 2


@pytest.mark.parametrize("numspecs", [1, 2, 5])
def test_cmd_spectrum_activate(numspecs):
    assert len(s.spectra.dict) == 0
    for _ in range(numspecs):
        hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == numspecs

    for specid in range(numspecs):
        hdtvcmd("spectrum activate {}".format(specid))
        f, ferr = hdtvcmd("spectrum list")
        for i, line in enumerate(f.split("\n")):
            if i == specid + 2:
                assert "|   AV |" in line
            elif i > 2:
                assert "|    V |" in line


@pytest.mark.parametrize("numspecs", [0, 1, 2, 5])
def test_cmd_spectrum_activate_multi(numspecs):
    assert len(s.spectra.dict) == 0
    for _ in range(numspecs):
        hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == numspecs

    f, ferr = hdtvcmd("spectrum activate all")
    if numspecs > 1:
        assert "Can only activate one spectrum" in ferr
    else:
        assert "" == ferr
        assert "" == f


def test_cmd_spectrum_info():
    assert len(s.spectra.dict) == 0
    hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == 1

    f, ferr = hdtvcmd("spectrum info")
    assert ferr == ""
    # If lots of stuff is printed, `spectrum info` is doing its job.
    assert len(f.split("\n")) == 9


@pytest.mark.parametrize("name", ["myname"])
def test_cmd_spectrum_name(name):
    assert len(s.spectra.dict) == 0
    hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == 1

    f, ferr = hdtvcmd("spectrum name {}".format(name))
    print(ferr)
    print(f)
    assert ferr == ""
    assert "Renamed spectrum 0 to &#x27;{}&#x27;".format(name) in f

    f, ferr = hdtvcmd("spectrum list")
    assert name in f


def test_cmd_spectrum_normalize():
    assert len(s.spectra.dict) == 0
    for _ in range(3):
        hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == 3
    for specids, norm in [("0", 0.3), ("0 1 2", 0.2)]:
        hdtvcmd("spectrum normalize {} {}".format(specids, norm))
        assert get_spec(0).norm == norm
    assert get_spec(2).norm == 0.2
    # TODO: Is there anything else supposed to happen when normalizing?


@pytest.mark.parametrize("ngroup", [2, 20])
def test_cmd_spectrum_rebin(ngroup):
    assert len(s.spectra.dict) == 0
    hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == 1

    with warnings.catch_warnings(record=True) as w:
        f, ferr = hdtvcmd("spectrum rebin 0 {}".format(ngroup))
        assert "Rebinning 0 with {} bins per new bin".format(ngroup) in f
        assert get_spec(0).hist.hist.GetNbinsX() == 8192 // ngroup


@pytest.mark.parametrize("binsize", [1, 0.5, 1.1])
def test_cmd_spectrum_calbin_binsize(binsize):
    assert len(s.spectra.dict) == 0
    hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == 1

    with warnings.catch_warnings(record=True) as w:
        f, ferr = hdtvcmd("spectrum calbin -b {} 0".format(binsize))
        assert "Calbinning 0 with binsize={}".format(binsize) in f
        assert get_spec(0).hist.hist.GetNbinsX() == int(ceil(8191 / binsize)) + 1


@pytest.mark.parametrize("parameter", ["-t", "-r"])
def test_cmd_spectrum_calbin_root_tv(parameter):
    assert len(s.spectra.dict) == 0
    hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == 1

    with warnings.catch_warnings(record=True) as w:
        f, ferr = hdtvcmd("spectrum calbin {} 0".format(parameter))
        assert "Calbinning 0 with binsize=" in f


@pytest.mark.parametrize("spline_order", [1, 3, 5])
def test_cmd_spectrum_calbin_spline_order(spline_order):
    assert len(s.spectra.dict) == 0
    hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == 1

    with warnings.catch_warnings(record=True) as w:
        f, ferr = hdtvcmd("spectrum calbin -k {} 0".format(spline_order))
        assert "Calbinning 0 with" in f


def test_cmd_spectrum_calbin_calibrated():
    assert len(s.spectra.dict) == 0
    hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == 1
    hdtvcmd("calibration position set 5 2.2".format(testspectrum))

    with warnings.catch_warnings(record=True) as w:
        f, ferr = hdtvcmd("spectrum calbin -d 0")
        assert "Calbinning 0 with" in f


def test_cmd_spectrum_resample():
    assert len(s.spectra.dict) == 0
    hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == 1

    with warnings.catch_warnings(record=True) as w:
        f, ferr = hdtvcmd("spectrum resample 0")
        assert "Total area changed by" in f


@pytest.mark.parametrize(
    "copy, matches",
    [
        ("0", ["0 to 3"]),
        ("0 1", ["0 to 3", "1 to 4"]),
        ("2 -s 3000", ["2 to 3000"]),
        ("0,1 -s 8,7", ["0 to 7", "1 to 8"]),
        ("0-2", ["0 to 3", "1 to 4", "2 to 5"]),
        ("2 -s 0", ["2 to 0"]),
    ],
)
def test_cmd_spectrum_copy(copy, matches):
    assert len(s.spectra.dict) == 0
    for _ in range(3):
        hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == 3
    f, ferr = hdtvcmd("spectrum copy {}".format(copy))
    for match in matches:
        assert "Copied spectrum {}".format(match) in f


def test_cmd_spectrum_copy_nonexistant():
    assert len(s.spectra.dict) == 0
    f, ferr = hdtvcmd("spectrum copy 0")
    assert "WARNING: Non-existent id 0" in ferr
    assert "WARNING: Nothing to do" in ferr


@pytest.mark.parametrize("fmt", ["lc", "txt"])
def test_cmd_spectrum_write(fmt, temp_file, suffix=""):
    assert len(s.spectra.dict) == 0
    hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == 1
    f, ferr = hdtvcmd("spectrum write {} {}{}".format(temp_file, fmt, suffix))
    assert "Wrote spectrum with id 0 to file {}".format(temp_file) in f
    assert os.path.getsize(temp_file) > 0


@pytest.mark.parametrize("fmt", ["lc", "txt"])
def test_cmd_spectrum_write_withid(fmt, temp_file):
    test_cmd_spectrum_write(fmt, temp_file, suffix=" 0")


def test_cmd_spectrum_write_invalid_format(temp_file):
    assert len(s.spectra.dict) == 0
    hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == 1
    f, ferr = hdtvcmd("spectrum write {} invalid 0".format(temp_file))
    assert os.path.getsize(temp_file) == 0
    assert "Failed to write spectrum: Invalid format specified" in ferr


@pytest.mark.parametrize(
    "factor, specids, response",
    [
        (0.1, "0", ["0"]),
        (0.2, "0,1", ["0", "1"]),
        (0.5, "0 2", ["0", "2"]),
        (20, "0-2", ["0", "1", "2"]),
        (2, "0", ["0"]),
    ],
)
def test_cmd_spectrum_multiply(factor, specids, response):
    assert len(s.spectra.dict) == 0
    for _ in range(3):
        hdtvcmd("spectrum get {}".format(testspectrum))
    assert len(s.spectra.dict) == 3
    orig_count = get_spec(0).hist.hist.GetSum()
    f, ferr = hdtvcmd("spectrum multiply {} {}".format(specids, factor))
    for spec in response:
        assert "Multiplying {} with {}".format(spec, factor) in f
    new_count = get_spec(0).hist.hist.GetSum()
    assert orig_count * factor - new_count < 0.00001 * new_count


@pytest.mark.parametrize("specfile0, specfile1", [(testspectrum, testspectrum)])
def test_cmd_spectrum_add(specfile0, specfile1):
    assert len(s.spectra.dict) == 0
    hdtvcmd("spectrum get {}".format(specfile0))
    hdtvcmd("spectrum get {}".format(specfile1))
    assert len(s.spectra.dict) == 2
    spec0_count = get_spec(0).hist.hist.GetSum()
    spec1_count = get_spec(1).hist.hist.GetSum()
    f, ferr = hdtvcmd("spectrum add 2 0 1")
    assert "Adding 0 to 2" in f
    spec2_count = get_spec(2).hist.hist.GetSum()
    assert spec0_count + spec1_count - spec2_count < 0.00001 * spec2_count


@pytest.mark.parametrize("specfile0, specfile1", [(testspectrum, testspectrum)])
def test_cmd_spectrum_add_normalize(specfile0, specfile1):
    assert len(s.spectra.dict) == 0
    hdtvcmd("spectrum get {}".format(specfile0))
    hdtvcmd("spectrum get {}".format(specfile1))
    assert len(s.spectra.dict) == 2
    spec0_count = get_spec(0).hist.hist.GetSum()
    spec1_count = get_spec(1).hist.hist.GetSum()
    f, ferr = hdtvcmd("spectrum add -n 2 0 1")
    assert "Adding 0 to 2" in f
    spec2_count = get_spec(2).hist.hist.GetSum()
    assert spec0_count + spec1_count - 2 * spec2_count < 0.00001 * spec2_count


@pytest.mark.parametrize("specfile0, specfile1", [(testspectrum, testspectrum)])
def test_cmd_spectrum_subtract(specfile0, specfile1):
    assert len(s.spectra.dict) == 0
    hdtvcmd("spectrum get {}".format(specfile0))
    hdtvcmd("spectrum get {}".format(specfile1))
    assert len(s.spectra.dict) == 2
    spec0_count = get_spec(0).hist.hist.GetSum()
    spec1_count = get_spec(1).hist.hist.GetSum()
    f, ferr = hdtvcmd("spectrum subtract 0 1")
    assert "Subtracting 1 from 0" in f
    new_count = get_spec(0).hist.hist.GetSum()
    assert spec0_count - spec1_count - new_count < 0.00001 * spec0_count


@pytest.mark.skip(reason="What is spectrum update supposed to do?")
def test_cmd_spectrum_update():
    pass


def get_spec(specid):
    return s.spectra.dict.get([x for x in list(s.spectra.dict) if x.major == specid][0])



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Preliminary ROOT file interface for hdtv
"""

import os
import fnmatch

import ROOT
import hdtv.rootext.display

import hdtv.cmdline
import hdtv.tabformat
import hdtv.rfile_utils
import hdtv.util
import hdtv.ui

from hdtv.spectrum import Spectrum
from hdtv.histogram import Histogram, RHisto2D, THnSparseWrapper
from hdtv.matrix import Matrix


class RootFileInterface(object):
    def __init__(self, spectra):
        hdtv.ui.debug("Loaded user interface for working with root files")

        self.spectra = spectra
        self.window = spectra.window
        self.caldict = spectra.caldict
        self.rootfile = None
        self.browser = None
        self.matviews = list()

        hdtv.cmdline.AddCommand("root ls", self.RootLs, maxargs=1)
        hdtv.cmdline.AddCommand("root ll", self.RootLL, maxargs=1, level=0)
        hdtv.cmdline.AddCommand("root pwd", self.RootPwd, nargs=0)
        hdtv.cmdline.AddCommand("root browse", self.RootBrowse, nargs=0)
        hdtv.cmdline.AddCommand("root open", self.RootOpen, nargs=1, fileargs=True)
        hdtv.cmdline.AddCommand("root close", self.RootClose, nargs=0)
        hdtv.cmdline.AddCommand(
            "root cd",
            self.RootCd,
            nargs=1,
            completer=hdtv.rfile_utils.RootCdCompleter(),
        )

        prog = "root get"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog)
        parser.add_argument(
            "-r",
            "--replace",
            action="store_true",
            default=False,
            help="replace existing histogram list",
        )
        parser.add_argument(
            "-c",
            "--load-cal",
            action="store_true",
            default=False,
            help="load calibration from calibration dictionary",
        )
        parser.add_argument(
            "-v",
            "--invisible",
            action="store_true",
            default=False,
            help="do not make histograms visible, only add to display list",
        )
        parser.add_argument("pattern", nargs="+")
        hdtv.cmdline.AddCommand(
            "root get",
            self.RootGet,
            completer=hdtv.rfile_utils.RootCompleter(),
            parser=parser,
        )

        # FIXME: make use of matrix ID possible for already loaded matrix
        prog = "root matrix view"
        description = "show a 2D view of the matrix"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument("matname", nargs="+")
        hdtv.cmdline.AddCommand(
            prog,
            self.RootMatrixView,
            completer=hdtv.rfile_utils.RootCompleter(),
            parser=parser,
        )

        prog = "root matrix get"
        description = "load a matrix, i.e. the projections, from a ROOT file.\n"
        description += "If the matrix is symmetric it only loads one projection.\n"
        description += "If it is asymmetric both projections will be loaded."
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default=None,
            help="base id for loaded projections",
        )
        parser.add_argument("matrix_type", metavar="matrix-type", help="{asym,sym}")
        parser.add_argument(
            "filename", metavar="matrix-file", help="file with matrix to load"
        )
        hdtv.cmdline.AddCommand(
            prog,
            self.RootMatrixGet,
            completer=hdtv.rfile_utils.RootCompleter(),
            parser=parser,
        )

        prog = "root cut view"
        description = "load a cut (TCutG) from a ROOT file and display it"
        description += "overlaid on the current matrix view"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-i",
            "--invert-axes",
            action="store_true",
            default=False,
            help="Exchange coordinate axes of cut (x <-> y)",
        )
        parser.add_argument("path", nargs="+", help="path of root cut")
        hdtv.cmdline.AddCommand(
            prog,
            self.RootCutView,
            completer=hdtv.rfile_utils.RootCompleter(),
            parser=parser,
        )

        prog = "root cut delete"
        description = "delete all cuts currently shown"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        hdtv.cmdline.AddCommand(prog, self.RootCutDelete, parser=parser)

        hdtv.cmdline.RegisterInteractive("gRootFile", self.rootfile)

    def RootBrowse(self, args):
        self.browser = ROOT.TBrowser()
        if self.rootfile is not None and not self.rootfile.IsZombie():
            self.rootfile.Browse(self.browser)

    def RootLs(self, args):
        keys = ROOT.gDirectory.GetListOfKeys()
        if not keys:
            return

        names = [k.GetName() for k in keys]

        if len(args) > 0:
            names = fnmatch.filter(names, args[0])

        names.sort()
        hdtv.tabformat.tabformat(names)

    def RootLL(self, args):
        if len(args) == 0:
            ROOT.gDirectory.ls()
        else:
            ROOT.gDirectory.ls(args[0])

    def RootCd(self, args):
        try:
            target = args[0]
        except IndexError:
            target = ""
        if self.rootfile:
            cur_root_dir = ROOT.gDirectory
        else:
            cur_root_dir = None

        (posix_path, rfile, root_dir) = hdtv.rfile_utils.GetRelDirectory(
            os.getcwd(), cur_root_dir, target
        )
        if (posix_path, rfile, root_dir) == (None, None, None):
            hdtv.ui.error("Invalid path specified.")
            return

        os.environ["OLDPWD"] = os.getcwd()
        os.chdir(posix_path)

        # If root_dir is None, we moved outside the ROOT file and are now in
        # a POSIX directory. If rfile is not None, we *changed* the ROOT file.
        if root_dir is None or rfile is not None:
            if self.rootfile is not None:
                hdtv.ui.info("Closing old root file %s" % self.rootfile.GetName())
                self.rootfile.Close()
            self.rootfile = rfile
            if self.rootfile is not None:
                hdtv.ui.info("Opened new root file %s" % self.rootfile.GetName())

        if root_dir is not None:
            root_dir.cd()

    def RootPwd(self, args):
        if self.rootfile is not None:
            ROOT.gDirectory.pwd()
        else:
            hdtv.ui.msg(os.getcwd())

    def RootOpen(self, args):
        return self.RootCd(args)

    def RootClose(self, args):
        if self.rootfile is None:
            hdtv.ui.info("No root file open, no action taken")
        else:
            self.rootfile.Close()
            self.rootfile = None

    def GetObj(self, path):
        """
        Load an object from a ROOT file
        """

        dirs = path.rsplit("/", 1)
        if len(dirs) == 1:
            # Load object from current directory
            obj = ROOT.gDirectory.Get(dirs[0])
        else:
            if self.rootfile:
                cur_root_dir = ROOT.gDirectory
            else:
                cur_root_dir = None

            (posix_path, rfile, root_dir) = hdtv.rfile_utils.GetRelDirectory(
                os.getcwd(), cur_root_dir, dirs[0]
            )

            if root_dir:
                obj = root_dir.Get(dirs[1])
            else:
                obj = None

            if rfile:
                rfile.Close()

        return obj

    def GetCut(self, path):
        """
        Load a cut (TCutG) from a ROOT file.
        """
        cut = self.GetObj(path)

        if cut is None or not isinstance(cut, ROOT.TCutG):
            hdtv.ui.error("%s is not a ROOT cut (TCutG)" % path)
            return None

        return cut

    def GetTH2(self, path):
        """
        Load a 2D histogram (``matrix'') from a ROOT file. This can be either
        a ``true'' TH2 histogram or a THnSparse of dimension 2.

        Note: Unlike RootGet(), this function does not support shell-style
        pattern expansion, as this make it too easy to undeliberately load
        too many histograms. As they use a lot of memory, this would likely
        lead to a crash of the program.
        """
        hist = self.GetObj(path)

        if hist is None or not (
            isinstance(hist, ROOT.TH2)
            or (isinstance(hist, ROOT.THnSparse) and hist.GetNdimensions() == 2)
        ):
            hdtv.ui.error("%s is not a 2d histogram" % path)
            return None

        return hist

    def RootCutView(self, args):
        """
        Load a ROOT cut (TCutG) from a ROOT file and display it.
        """
        if len(self.matviews) == 0:
            raise hdtv.cmdline.HDTVCommandError(
                "Cannot display cut: no matrix view open"
            )
        for path in args.path:
            cut = self.GetCut(path)
            if cut:
                self.matviews[-1].AddCut(cut, args.invert_axes)

    def RootCutDelete(self, args):
        """
        Delete all cuts shown in the current matrix view.
        """
        if len(self.matviews) == 0:
            raise hdtv.cmdline.HDTVCommandError("No matrix view open")
        self.matviews[-1].DeleteAllCuts()

    def RootMatrixView(self, args):
        """
        Load a 2D histogram (``matrix'') from a ROOT file and display it.
        """
        for path in args.matname:
            hist = self.GetTH2(path)
            if hist:
                title = hist.GetTitle()
                viewer = ROOT.HDTV.Display.MTViewer(400, 400, hist, title)
                self.matviews.append(viewer)

    def RootMatrixGet(self, args):
        """
        Load a 2D histogram (``matrix'') from a ROOT file in projection mode.
        """
        # FIXME: Copy and paste from matInterface.py -> unify?
        if args.spectrum is not None:
            ID = args.spectrum
        else:
            ID = None

        if args.matrix_type == "sym":
            sym = True
        elif args.matrix_type == "asym":
            sym = False
        else:
            # FIXME: is there really no way to test that automatically????
            raise hdtv.cmdline.HDTVCommandError(
                "Please specify if matrix is of type asym or sym"
            )
        rhist = self.GetTH2(args.filename)
        if rhist is None:
            raise hdtv.cmdline.HDTVCommandError("Failed to open 2D histogram")

        if isinstance(rhist, ROOT.TH2):
            hist = RHisto2D(rhist)
        elif isinstance(rhist, ROOT.THnSparse):
            hist = RHisto2D(THnSparseWrapper(rhist))
        else:
            raise RuntimeError

        matrix = Matrix(hist, sym, self.spectra.viewport)
        ID = self.spectra.GetFreeID()
        matrix.ID = ID
        matrix.color = hdtv.color.ColorForID(ID.major)
        # load x projection
        proj = matrix.xproj
        sid = self.spectra.Insert(proj, ID=hdtv.util.ID(ID.major, 0))
        self.spectra.ActivateObject(sid)
        # for asym matrix load also y projection
        if sym is False:
            proj = matrix.yproj
            self.spectra.Insert(proj, ID=hdtv.util.ID(ID.major, 1))

    def RootGet(self, args):
        """
        Load spectra from Rootfile
        """
        if self.rootfile:
            cur_root_dir = ROOT.gDirectory
        else:
            cur_root_dir = None

        objs = []
        for pat in args.pattern:
            objs += hdtv.rfile_utils.Get(".", cur_root_dir, pat)

        if args.replace:
            self.spectra.Clear()

        loaded = list()
        if self.window.viewport:
            self.window.viewport.LockUpdate()
        try:  # We should really use a context manager here...
            for obj in objs:
                if isinstance(obj, ROOT.TH1):
                    spec = Spectrum(Histogram(obj))
                    sid = self.spectra.Insert(spec)
                    spec.color = hdtv.color.ColorForID(sid.major)
                    loaded.append(sid)
                    if spec.cal:
                        self.caldict[spec.name] = spec.cal
                    else:
                        if args.load_cal:
                            if spec.name in self.spectra.caldict:
                                spec.cal = self.caldict[spec.name]
                            else:
                                hdtv.ui.warning(
                                    "No calibration found for %s" % spec.name
                                )
                else:
                    hdtv.ui.warning("%s is not a 1D histogram object" % obj.GetName())
            hdtv.ui.msg("%d spectra loaded" % len(loaded))
            if args.invisible:
                self.spectra.HideObjects(loaded)
            elif len(loaded) > 0:
                # activate last loaded spectrum
                self.spectra.ActivateObject(loaded[-1])
            # Expand window if it is the only spectrum
            if len(self.spectra) == 1:
                self.window.Expand()
        finally:
            if self.window.viewport:
                self.window.viewport.UnlockUpdate()


# plugin initialisation
import __main__

r = RootFileInterface(__main__.spectra)



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import os
import hdtv.ui
import hdtv.util
import hdtv.cmdline

preamble = r"""\makeatletter
\@ifundefined{standalonetrue}{\newif\ifstandalone}{\let\ifbackup=\ifstandalone}
\@ifundefined{section}{\standalonetrue}{\standalonefalse}
\makeatother
\ifstandalone

\documentclass[12pt,oneside,a4paper]{report}
\usepackage[ngerman]{babel}
\usepackage{amsmath}
\usepackage{amssymb}
\usepackage{graphicx}
\usepackage{a4}
\usepackage[T1]{fontenc}
\usepackage{ae,aecompl}
\usepackage[utf8]{inputenc}
\usepackage[font=small,labelfont=bf,textfont=it]{caption}
\usepackage{longtable}


\begin{document}
\fi
\begin{center}
\begin{longtable}"""

enddok = r"""\end{longtable}
\end{center}

\ifstandalone
\end{document}
\else
\let\ifstandalone=\ifbackup
\expandafter\endinput
\fi
"""


class TexTable(hdtv.util.Table):
    def __init__(self, data, keys, header=None, sortBy=None, reverseSort=False, ha="c"):
        hdtv.util.Table.__init__(
            self,
            data,
            keys,
            header,
            ignoreEmptyCols=False,
            sortBy=sortBy,
            reverseSort=reverseSort,
            extra_header=preamble,
            extra_footer=enddok,
        )
        self.col_sep_char = "&"
        self.empty_field = " "
        self.ha = ha

    def build_header(self):
        header = r"\hline" + os.linesep
        for i in range(2):
            for string in self.header:
                string = string.replace("_", "-")
                header += r"\multicolumn{1}{%s}{\textbf{%s}} &" % (self.ha, string)
            header = header.rstrip("&")
            header += r"\\ \hline" + os.linesep
            if i == 0:
                header += r"\endfirsthead" + os.linesep + os.linesep
                header += r"\hline" + os.linesep
            else:
                header += r"\endhead" + os.linesep + os.linesep

        header += r"\hline" + os.linesep
        header += (
            r"\multicolumn{%d}{r}{{wird fortgesetzt...}} \\\ " % len(self.header)
            + os.linesep
        )
        header += r"\endfoot" + os.linesep + os.linesep

        header += r"\hline " + os.linesep
        header += r"\endlastfoot" + os.linesep + os.linesep

        return header

    def __str__(self):
        text = str()
        if self.extra_header is not None:
            text += (
                str(self.extra_header)
                + "{"
                + len(self.keys) * self.ha
                + "}"
                + os.linesep
            )

        lines = self.build_lines()

        header_line = self.build_header()
        text += header_line + os.linesep

        # Build lines
        for line in lines:
            line_str = ""
            for col in range(0, len(line)):
                if not self._ignore_col[col]:
                    line_str += str(" " + line[col] + " ").rjust(self._col_width[col])
                if col != len(line) - 1:
                    line_str += self.col_sep_char

            text += line_str + r"\\ " + os.linesep

        if self.extra_footer is not None:
            text += str(self.extra_footer)

        return text


class fitTex:
    """
    Create a table for use with latex.

    The resulting file can be compiled standalone or included in another latex
    document. It uses the longtable package.
    """

    def __init__(self, spectra, fitInterface):
        hdtv.ui.debug("Loaded plugin for exporting fitlist to tex")

        self.spectra = spectra
        self.f = fitInterface

        prog = "fit tex"
        description = "create a table in latex format"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-c",
            "--columns",
            action="store",
            default=None,
            help="values to include as columns of the table",
        )
        parser.add_argument(
            "-H", "--header", action="store", default=None, help="header of columns"
        )
        parser.add_argument(
            "-k", "--key-sort", action="store", default=None, help="sort by key"
        )
        parser.add_argument(
            "-r",
            "--reverse-sort",
            action="store_true",
            default=False,
            help="reverse the sort",
        )
        parser.add_argument(
            "-a",
            "--alignment",
            action="store",
            default="c",
            help="horizontal alignment (c=center, l=left or r=right)",
        )
        parser.add_argument(
            "-f",
            "--fit",
            action="store",
            default="all",
            help="specify fits to include in table",
        )
        parser.add_argument(
            "-s", "--spectrum", action="store", default="active", help="specify spectra"
        )
        parser.add_argument("filename", metavar="output-file", help="file to write to")
        hdtv.cmdline.AddCommand(prog, self.WriteTex, parser=parser)

    def WriteTex(self, args):
        filename = os.path.expanduser(args.filename)

        # get list of fits
        fits = list()
        sids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        if len(sids) == 0:
            hdtv.ui.warning("No spectra chosen or active")
            return
        for sid in sids:
            spec = self.spectra.dict[sid]
            ids = hdtv.util.ID.ParseIds(args.fit, spec)
            fits.extend([spec.dict[ID] for ID in ids])
        (peaklist, params) = self.f.ExtractFits(fits)

        # keys
        if args.columns is not None:
            keys = args.columns.split(",")
        else:
            keys = params

        # header
        if args.header is not None:
            header = args.header.split(",")
        else:
            header = None

        # sort key
        if args.key_sort:
            sortBy = args.key_sort.lower()
        else:
            sortBy = keys[0]

        # alignment
        if args.alignment.lower() in ["c", "l", "r"]:
            ha = args.alignment.lower()
        else:
            msg = "Invalid specifier for horizontal alignment. "
            msg += "Valid values are c,l,r. "
            hdtv.ui.warning(msg)
            ha = "c"

        # do the work
        table = TexTable(peaklist, keys, header, sortBy, args.reverse_sort, ha)
        with open(filename, "w") as out:
            out.write(str(table))


# plugin initialisation
import __main__
from hdtv.plugins.fitInterface import fit_interface

hdtv.cmdline.RegisterInteractive("fittex", fitTex(__main__.spectra, fit_interface))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function

from hdtv.ui import *
import hdtv.database

db = hdtv.database.PGAALibraries.PGAAlib_IKI2000()

output = db.find(energy=511, sort_key="k0", fuzziness=2)

table = Table(
    output,
    header=["ID", "Energy / (keV)", "Intensity / (%)", "Sigma / (b)"],
    attrs=["id", "energy", "intensity", "sigma"],
    sortBy="intensity",
    reverseSort=True,
)

table.out()


table2 = Table(output, attrs=["id", "energy", "intensity", "sigma"])
print(table2)



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import ROOT
from .background import BackgroundModel


class BackgroundModelInterpolation(BackgroundModel):
    """
    Interpolation background model
    """

    def __init__(self):
        super(BackgroundModelInterpolation, self).__init__()
        self.fParStatus = {"nparams": 3}
        self.fValidParStatus = {"nparams": [int]}

        self.ResetParamStatus()
        self.name = "interpolation"
        self.requiredBgRegions = 3

    def ResetParamStatus(self):
        """
        Reset parameter status to defaults
        """
        self.fParStatus["nparams"] = 3

    def GetFitter(self, integrate, likelihood, nparams=None, nbg=None):
        """
        Creates a C++ Fitter object, which can then do the real work
        integrate and likelihood are ignored (do not make sense here)
        """

        if nbg is not None:
            if nbg < self.requiredBgRegions:
                return None
            self.fFitter = ROOT.HDTV.Fit.InterpolationBg(2 * nbg)
            self.fParStatus["nparams"] = 2 * nbg
        elif nparams is not None:
            self.fFitter = ROOT.HDTV.Fit.InterpolationBg(nparams)
            self.fParStatus["nparams"] = nparams
        else:
            msg = (
                "Status specifier %s of background fitter is invalid."
                % fParStatus["nparams"]
            )
            raise ValueError(msg)

        self.ResetGlobalParams()

        return self.fFitter



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

# -------------------------------------------------------------------------
# Infrastructure for generation of artificial spectra
# -------------------------------------------------------------------------

from __future__ import division

import math
import ROOT


class PolyBg:
    "Polynomial background"

    def __init__(self, coeff):
        self.coeff = coeff

    def value(self, x):
        y = 0.0
        for c in reversed(self.coeff):
            y = y * x + c
        return y


class TheuerkaufPeak:
    "Classic TV peak (a.k.a. Theuerkauf peak)"

    def __init__(self, pos, vol, fwhm, tl=None, tr=None, sh=None, sw=None):
        self.pos = pos
        self.sigma = fwhm / (2 * math.sqrt(2 * math.log(2)))
        self.tl = tl
        self.tr = tr
        self.sh = sh
        self.sw = sw

        # Calculate normalization
        # Contribution from left tail and left half of truncated Gaussian
        if self.tl:
            norm = (
                self.sigma ** 2
                / self.tl
                * math.exp(-self.tl ** 2 / (2 * self.sigma ** 2))
            )
            norm += (
                math.sqrt(math.pi / 2)
                * self.sigma
                * ROOT.TMath.Erf(self.tl / (math.sqrt(2) * self.sigma))
            )
        else:
            norm = math.sqrt(math.pi / 2) * self.sigma

        # Contribution from right tail and right half of truncated Gaussian
        if self.tr:
            norm += (
                self.sigma ** 2
                / self.tr
                * math.exp(-self.tr ** 2 / (2 * self.sigma ** 2))
            )
            norm += (
                math.sqrt(math.pi / 2)
                * self.sigma
                * ROOT.TMath.Erf(self.tr / (math.sqrt(2) * self.sigma))
            )
        else:
            norm += math.sqrt(math.pi / 2) * self.sigma

        # Calculate amplitude
        self.amp = vol / norm

    def step(self, x):
        if self.sw is not None and self.sh is not None:
            dx = x - self.pos
            _y = math.pi / 2.0 + math.atan(self.sw * dx / (math.sqrt(2) * self.sigma))
            return self.sh * _y
        else:
            return 0.0

    def value(self, x):
        dx = x - self.pos
        if self.tl is not None and dx < -self.tl:
            _y = self.tl / (self.sigma ** 2) * (dx + self.tl / 2)
        elif self.tr is not None and dx > self.tr:
            _y = -self.tr / (self.sigma ** 2) * (dx - self.tr / 2)
        else:
            _y = -(dx ** 2) / (2 * self.sigma ** 2)

        return self.amp * (math.exp(_y) + self.step(x))


class EEPeak:
    "Electron-electron scattering peak"

    def __init__(self, pos, amp, sigma1, sigma2, eta, gamma):
        self.pos = pos
        self.amp = amp
        self.sigma1 = sigma1
        self.sigma2 = sigma2
        self.eta = eta
        self.gamma = gamma

    def value(self, x):
        dx = x - self.pos
        if dx <= 0:
            _y = math.exp(-math.log(2) * dx ** 2 / self.sigma1 ** 2)
        elif dx <= self.eta * self.sigma2:
            _y = math.exp(-math.log(2) * dx ** 2 / self.sigma2 ** 2)
        else:
            B = self.sigma2 * self.gamma - 2.0 * self.sigma2 * self.eta ** 2 * math.log(
                2
            )
            B /= 2.0 * self.eta * math.log(2)
            A = 2 ** (-self.eta ** 2) * (self.sigma2 * self.eta + B) ** self.gamma
            _y = A / (B + dx) ** self.gamma

        return _y * self.amp


class SpecFunc:
    """
    Sum of background and (several) peak functions, describing a complete
    spectrum
    """

    def __init__(self):
        self.background = None
        self.peaks = []

    def __call__(self, x):
        y = 0.0

        if self.background:
            y += self.background.value(x[0])

        for peak in self.peaks:
            y += peak.value(x[0])

        return y


class Spectrum:
    "Class to generate artificial spectra"

    def __init__(self, nbins, xmin, xmax):
        self.func = SpecFunc()
        self.nbins = nbins
        self.xmin = xmin
        self.xmax = xmax

    def GetFunc(self, name="specfunc"):
        """
        Return the underlying function as a ROOT TF1 object.
        name is the name to be used by ROOT.
        """
        return ROOT.TF1(name, self.func, self.xmin, self.xmax, 0)

    def GetExactHist(self, name="spec_exact"):
        """
        Return a ROOT TH1D object with each bin containing exactly the value
        of the spectrum function at its center.
        """
        hist = ROOT.TH1D(name, name, self.nbins, self.xmin, self.xmax)

        bincenter = hist.GetXaxis().GetBinCenter
        for bin in range(1, hist.GetNbinsX() + 1):
            hist.SetBinContent(bin, self.func([bincenter(bin)]))

        return hist

    def GetSampledHist(self, nsamples, name="spec_sampled"):
        """
        Return a ROOT TH1I object filled with nsamples samples from the spectrum
        function.
        """
        # Due to limitations in ROOT, we cannot pass a TF1 object to FillRandom(),
        # but the function must be loaded from the ROOT function dictionary.
        # We need to choose a "special" name in order to not interfere with
        # any functions the user may have defined.
        func = self.GetFunc("__samplefunc__")

        hist = ROOT.TH1D(name, name, self.nbins, self.xmin, self.xmax)
        hist.FillRandom("__samplefunc__", nsamples)

        del func

        return hist


def write_hist(hist, fname, include_x=False, include_err=False):
    """
    Dump a ROOT histogram object to a textfile. We write the content of all
    bins, exculding the under- and overflow bins, seperated by newlines. The
    parameters include_x and include_err control if the bin center and the bin
    error are to be included. If the resulting file should be readable by the
    old tv program, they must both be set to False.
    """
    f = open(fname, "w")

    axis = hist.GetXaxis()
    for bin in range(1, hist.GetNbinsX() + 1):
        if include_x:
            f.write("%f " % axis.GetBinCenter(bin))
        f.write("%f" % hist.GetBinContent(bin))
        if include_err:
            f.write(" %f" % hist.GetBinError(bin))
        f.write("\n")

    f.close()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2009 Dirk Meyer, Jason Tackaberry
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""
This file contains a wrapper for weakref that the weak reference can
be used the same way the real object would be used.
"""

__all__ = ["weakref"]

# python imports
from _weakref import ref


class NoneProxy(object):
    def __call__(self):
        return None


class weakref(object):
    """
    This class represents a weak reference based on the python
    module weakref. The difference between weakref.ref and this class
    is that you can access the ref without calling it first.
    E.g.: foo = weak(bar) and bar has the attribute x. With a normal
    weakref.ref you need to call foo().x to get, this class makes it
    possible to just use foo.x.

    All functions are passed to the real object behind the weakref. To
    check if the weakref is alive or not, you can compare the object with
    None. Do not use a simple if, because an object still alive can also
    be False (e.g. an empty list).
    """

    def __new__(cls, object):
        # Make an instance of callable weakref if the object is callable.
        # This way callable(weakref) == False if the underlying object isn't
        # callable.
        if object is None:
            return None
        if callable(object):
            cls = _callable_weakref
        return super(weakref, weakref).__new__(cls)

    def __init__(self, object):
        if object is not None:
            if isinstance(object, weakref):
                self._ref = object._ref
            else:
                self._ref = ref(object)
        else:
            self._ref = NoneProxy()

    def __getattribute__(self, attr):
        if attr == "__class__":
            return self._ref().__class__
        if attr == "_ref":
            return object.__getattribute__(self, attr)
        return getattr(self._ref(), attr)

    def __setattr__(self, attr, value):
        if attr == "_ref":
            return object.__setattr__(self, attr, value)
        return setattr(self._ref(), attr, value)

    def __delattr__(self, attr):
        return delattr(self._ref(), attr)

    def __getitem__(self, key):
        return self._ref()[key]

    def __delitem__(self, key):
        del self._ref()[key]

    def __setitem__(self, key, value):
        self._ref()[key] = value

    def __bool__(self):
        if self._ref():
            return 1
        else:
            return 0

    def __eq__(self, other):
        if isinstance(other, weakref):
            other = other._ref()
        return self._ref() == other

    def __ne__(self, other):
        if isinstance(other, weakref):
            other = other._ref()
        return self._ref() != other

    def __gt__(self, other):
        if isinstance(other, weakref):
            other = other._ref()
        return self._ref() > other

    def __lt__(self, other):
        if isinstance(other, weakref):
            other = other._ref()
        return self._ref() < other

    def __ge__(self, other):
        if isinstance(other, weakref):
            other = other._ref()
        return self._ref() >= other

    def __le__(self, other):
        if isinstance(other, weakref):
            other = other._ref()
        return self._ref() <= other

    def __str__(self):
        return "<weakref proxy; %s>" % str(self._ref())

    def __deepcopy__(self, memo):
        return weakref(self._ref())


class _callable_weakref(weakref):
    def __new__(cls, object):
        return super(weakref, weakref).__new__(cls)

    def __call__(self, *args, **kwargs):
        return self._ref()(*args, **kwargs)



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Gamma database integration for HDTV
"""

import re
from html import escape

import hdtv.plugins
import hdtv.cmdline
import hdtv.options
import hdtv.database
import hdtv.ui

import hdtv.fit


class Database(object):
    def __init__(self):
        hdtv.ui.debug("Loaded Database Lookup plugin")

        self.opt = dict()
        self.database = None

        # Register configuration variables for fit peakfind
        self.opt["db"] = hdtv.options.Option(
            default="PGAAlib_IKI2000",
            parse=hdtv.options.parse_choices(list(hdtv.database.databases.keys())),
            changeCallback=lambda x: self.Set(x),
        )  # default database
        hdtv.options.RegisterOption("database.db", self.opt["db"])

        # Automatically lookup fitted peaks in database
        self.opt["auto_lookup"] = hdtv.options.Option(
            default=False,
            parse=hdtv.options.parse_bool,
            changeCallback=lambda x: self.SetAutoLookup(x),
        )
        hdtv.options.RegisterOption("database.auto_lookup", self.opt["auto_lookup"])

        # Set default database
        #        hdtv.options.Reset("database.db")

        self.opt["fuzziness"] = hdtv.options.Option(
            default=1.0, parse=lambda x: float(x)
        )
        hdtv.options.RegisterOption(
            "database.fuzziness", self.opt["fuzziness"]
        )  # Lookup fuzziness

        self.opt["sort_key"] = hdtv.options.Option(default=None)
        hdtv.options.RegisterOption("database.sort_key", self.opt["sort_key"])

        self.opt["sort_reverse"] = hdtv.options.Option(
            default=False, parse=hdtv.options.parse_bool
        )
        hdtv.options.RegisterOption("database.sort_reverse", self.opt["sort_reverse"])

        # TODO: proper help
        prog = "db lookup"
        description = "Lookup database entry"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-f",
            "--fuzziness",
            type=float,
            default=None,  # Default is handled via hdtv.options.Option
            help="Fuzziness for database lookup",
        )
        parser.add_argument(
            "-k",
            "--sort-key",
            action="store",
            default=None,  # Default is handled via hdtv.options.Option
            help="Sort by key",
        )
        parser.add_argument(
            "-r",
            "--sort-reverse",
            action="store_true",
            default=None,
            help="Reverse sorting",
        )
        parser.add_argument("specs", nargs="+")
        hdtv.cmdline.AddCommand(prog, self.Lookup, parser=parser, fileargs=False)

        prog = "db list"
        description = "Show available databases"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        hdtv.cmdline.AddCommand(prog, self.List, parser=parser, fileargs=False)

        prog = "db set"
        description = "Set database"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument("database")
        hdtv.cmdline.AddCommand(
            prog,
            lambda args: hdtv.options.Set("database.db", args.database),
            parser=parser,
            fileargs=False,
        )

        prog = "db info"
        description = "Show info about database"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument("database", default=None, nargs="*")
        hdtv.cmdline.AddCommand(prog, self.Info, parser=parser, fileargs=False)

    def FitPeakPostHook(self, fitclass):
        """
        Hook for hdtv.fit.Fit.FitPeakFunc function to automatically list matching
        database entries
        """
        for p in fitclass.peaks:
            self.Lookup(["energy=" + str(p.pos_cal.value)], defaults=True)

    def SetAutoLookup(self, autolookup_opt):
        """
        Activate/Deactivate automatic lookup of peaks
        """
        if autolookup_opt.Get():
            if self.FitPeakPostHook not in hdtv.fit.Fit.FitPeakPostHooks:
                hdtv.fit.Fit.FitPeakPostHooks.append(self.FitPeakPostHook)
                hdtv.ui.msg("Autolookup activated")
        else:
            try:
                hdtv.fit.Fit.FitPeakPostHooks.remove(self.FitPeakPostHook)
            except ValueError:  # Ignore error for trying to remove nonexistant element
                pass
            else:
                hdtv.ui.msg("Autolookup deactivated")

    def Set(self, dbname=None, open=False):
        """
        Set database Callback

        open: If true automatically opens and reads database
        """
        if dbname is None:
            dbname = hdtv.options.Get("database.db")
        else:
            try:
                dbname = dbname.Get()  # dbname may be an option
            except AttributeError:
                pass
        try:
            db = hdtv.database.databases[dbname.lower()]
            if self.database != db:
                self.database = db()
                if open:
                    self.database.open()
                hdtv.ui.msg('"' + self.database.description + '" loaded')
                return True
        except KeyError:
            raise hdtv.cmdline.HDTVCommandError("No such database: " + dbname)

    def Info(self, args):
        """
        Print info about database(s)
        """
        if not args.database:
            args.database = [hdtv.options.Get("database.db")]

        for dbs in args.database:
            db = hdtv.database.databases[dbs.lower()]()
            try:
                hdtv.ui.msg(html=f"<b>Database</b>: {escape(db.name)}")
                hdtv.ui.msg(html=f"<b>Description</b>: {escape(db.description)}")
                self.showDBfields(db)

            except KeyError:
                raise hdtv.cmdline.HDTVCommandError("No such database: " + db.name)

    def assureOpen(self):
        """
        assure that the database has been opened
        """
        try:
            if not self.database.opened:
                self.database.open()
        except AttributeError:
            self.Set(open=True)

    def List(self, args):
        """
        List available databases
        """
        for (name, db) in list(hdtv.database.databases.items()):
            hdtv.ui.msg(html=f"<b>{escape(name)}</b>: {escape(db().description)}")

    def showDBfields(self, db=None):
        """
        Show fields of database db. If db is None show field of current database
        """

        html = "<b>Valid fields</b>: "
        if db is None:
            db = self.database
        html += ", ".join([f"'{str(s)}'" for s in db.fParamConv.keys()])
        hdtv.ui.msg(html=html)

    def Lookup(self, args, defaults=False):
        """
        Lookup entry in database

        args.specs should be something like "<fieldname>=value"
        if <fieldname> is omitted it is assumed to be energy
        """

        self.assureOpen()

        lookupargs = dict()

        # Valid arguments
        vargs = list()
        for v in self.database.fParamConv.keys():
            vargs.append(v.lower())

        # parse arguments
        for a in args.specs:
            m = re.match(r"(.*)=(.*)", a)
            if m is not None:
                if m.group(1).lower() in vargs:
                    lookupargs[m.group(1)] = m.group(2)
                else:
                    self.showDBfields()
                    return False
            else:
                lookupargs["energy"] = float(a)
                continue

        if defaults or args.sort_key is None:
            lookupargs["sort_key"] = hdtv.options.Get("database.sort_key")
        else:
            lookupargs["sort_key"] = args.sort_key

        if defaults or args.sort_reverse is None:
            lookupargs["sort_reverse"] = hdtv.options.Get("database.sort_reverse")
        else:
            lookupargs["sort_reverse"] = args.sort_reverse

        if defaults or args.fuzziness is None:
            fuzziness = hdtv.options.Get("database.fuzziness")
        else:
            fuzziness = args.fuzziness

        try:
            results = self.database.find(fuzziness, **lookupargs)
        except AttributeError:
            return False

        if len(results) > 0:
            table = hdtv.util.Table(
                results,
                header=self.database.fOrderedHeader,
                keys=self.database.fParamConv.keys(),
            )
            hdtv.ui.msg(html=str(table))

        hdtv.ui.msg("Found " + str(len(results)) + " results")


# plugin initialisation
database = Database()
hdtv.cmdline.RegisterInteractive("database", database)



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import re

from prompt_toolkit.completion import WordCompleter

import hdtv.options
import hdtv.cmdline
import hdtv.util


def ConfigSet(args):
    try:
        hdtv.options.Set(args.variable, args.value)
    except KeyError:
        raise hdtv.cmdline.HDTVCommandAbort(args.variable + ": no such option")
    except ValueError as err:
        raise hdtv.cmdline.HDTVCommandAbort(
            "Invalid value (%s) for option %s. %s" % (args.value, args.variable, err)
        )


def ConfigShow(args):
    if args.variable:
        try:
            hdtv.ui.msg(html=hdtv.options.Show(args.variable))
        except KeyError:
            hdtv.ui.warning(args.variable + ": no such option")
    else:
        hdtv.ui.msg(html=hdtv.options.Str(), end="")


def ConfigReset(args):
    if args.variable:
        if args.variable == "all":
            hdtv.options.ResetAll()
            hdtv.ui.debug("All configuration variables were reset.")
        else:
            try:
                hdtv.options.Reset(args.variable)
                hdtv.ui.debug("Reset configuration variable " + args.variable)
            except KeyError:
                hdtv.ui.warning(args.variable + ": no such option")
    else:
        hdtv.ui.msg(hdtv.options.Str(), end="")


config_var_completer = WordCompleter(
    hdtv.options.OptionManager.__dict__.keys(),
    pattern=re.compile(r"^([^\s]+)"),
)

prog = "config set"
description = "Set a configuration variable"
parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
parser.add_argument("variable")
parser.add_argument("value")
hdtv.cmdline.AddCommand(
    prog, ConfigSet, level=2, parser=parser, completer=config_var_completer
)

prog = "config show"
description = "Show the configuration or a single configuration variable"
parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
parser.add_argument("variable", nargs="?", default=None)
hdtv.cmdline.AddCommand(
    prog, ConfigShow, level=1, parser=parser, completer=config_var_completer
)

prog = "config reset"
description = "Reset a single configuration variable"
parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
parser.add_argument(
    "variable",
    nargs="?",
    default=None,
    help="Name of the variable to reset. Use 'all' to reset all configuration variables.",
)
hdtv.cmdline.AddCommand(
    prog, ConfigReset, level=2, parser=parser, completer=config_var_completer
)

hdtv.ui.debug("Loaded user interface for configuration variables")



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import tempfile
import os
import sys

import pytest


@pytest.fixture(scope="function")
def temp_file(request):
    """
    pytest fixture that provides a temporary file for writing
    """
    filename = tempfile.mkstemp(prefix="hdtv_")[1]
    os.remove(filename)
    yield filename
    try:
        os.remove(filename)
    except OSError:
        pass


@pytest.fixture(
    scope="function",
    params=[
        ".gz",
        pytest.param(
            ".xz",
            marks=pytest.mark.skipif(
                sys.version_info < (3, 0), reason="no module lzma in python2"
            ),
        ),
        pytest.param(
            ".bz2",
            marks=pytest.mark.skipif(
                sys.version_info < (3, 0), reason="no module bz2 in python2"
            ),
        ),
        "",
    ],
)
def temp_file_compressed(request):
    """
    pytest fixture that provides a temporary file for writing
    """
    filename = tempfile.mkstemp(prefix="hdtv_", suffix=str(request.param))[1]
    os.remove(filename)
    yield filename
    try:
        os.remove(filename)
    except OSError:
        pass



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2019  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import os

import pytest

from numpy import arange, exp, log, ones, savetxt, sqrt
from numpy.random import poisson
import numpy as np
from scipy.stats import norm

# Set properties of the background spectrum
# The properties are implemented in a separate class
# to be able to access them before the actual spectrum has
# been created
class ArtificialSpecProp:
    def __init__(self):
        self.nsteps = 3
        self.bg_regions = [[0.1, 0.35], [0.65, 0.775], [0.775, 0.9], [0.9, 0.95]]
        self.peak_width = 0.03
        self.nbins_per_step = 300
        self.bg_per_bin = 10.0
        self.peak_volume = 500.0
        self.bg_type = ["constant", "exponential", "constant"]
        self.poisson_fluctuations = [False, False, True]


class ArtificialSpec:
    def __init__(self, path, prop=ArtificialSpecProp()):
        self.nsteps = prop.nsteps
        self.bg_regions = prop.bg_regions
        self.peak_width = prop.peak_width
        self.nbins_per_step = prop.nbins_per_step
        self.bg_per_bin = prop.bg_per_bin
        self.peak_volume = prop.peak_volume
        self.bg_type = prop.bg_type
        self.poisson_fluctuations = prop.poisson_fluctuations
        self.spectrum = ones(self.nsteps * self.nbins_per_step)
        self.filename = os.path.join(path, "test_spectrum_background.tv")

    def create(self):
        """Create the test spectrum (if not already done)"""

        np.random.seed(0)

        if not os.path.isfile(self.filename):
            # Create an array for the bin centers
            bins = arange(self.nsteps * self.nbins_per_step) + 0.5
            spectrum = ones(self.nsteps * self.nbins_per_step)
            # Create the steps with increasing height
            for i in range(self.nsteps):
                if self.bg_type[i] == "constant":
                    spectrum[
                        i * self.nbins_per_step : (i + 1) * self.nbins_per_step
                    ] = self.bg_per_bin * ones(self.nbins_per_step)
                elif self.bg_type[i] == "exponential":
                    spectrum[
                        i * self.nbins_per_step : (i + 1) * self.nbins_per_step
                    ] = exp(
                        log(self.bg_per_bin)
                        + bins[i * self.nbins_per_step] / (0.5 * self.nbins_per_step)
                        - bins[i * self.nbins_per_step : (i + 1) * self.nbins_per_step]
                        / (0.5 * self.nbins_per_step)
                    )

                if self.poisson_fluctuations[i]:
                    spectrum[
                        i * self.nbins_per_step : (i + 1) * self.nbins_per_step
                    ] = poisson(
                        spectrum[
                            i * self.nbins_per_step : (i + 1) * self.nbins_per_step
                        ]
                    )

                # The peaks are located exactly in the centers of the steps
                spectrum = spectrum + self.peak_volume * norm.pdf(
                    bins,
                    loc=i * self.nbins_per_step + 0.5 * self.nbins_per_step,
                    scale=self.peak_width * self.nbins_per_step,
                )

            savetxt(self.filename, spectrum)

        return self.filename



# -*- coding: utf-8 -*-
# IAEA database, International Atomic Energy Agency

import json
import hdtv.util
import hdtv.ui
from uncertainties import ufloat
from hdtv.database.common import *

# TODO: integrate this in a class?


def SearchNuclide(nuclide):
    # Get all data as dict from json file
    alldata = json.loads(open(os.path.join(hdtv.datadir, "IAEA.json")).read())

    # Select nuclide
    data = next(x for x in alldata if x["nuclide"] == nuclide)
    if not data:
        errorText = "There is no nuclide called " + nuclide + " in the table."
        raise hdtv.cmdline.HDTVCommandError(errorText)

    # Use ufloat to represent values with uncertainties
    transitions = [
        {
            "energy": ufloat(t["energy"], t["energy_uncertainty"]),
            "intensity": ufloat(t["intensity"], t["intensity_uncertainty"]),
        }
        for t in data["transitions"]
    ]
    data["transitions"] = transitions
    data["halflife"] = ufloat(data["halflife"], data["halflife_uncertainty"])
    del data["halflife_uncertainty"]

    return data



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from math import pow
from uncertainties.umath import log, exp
from .efficiency import _Efficiency
from ROOT import TF1, TF2
from hdtv.util import Pairs


class PolyEff(_Efficiency):
    """
    'Polynom' efficiency

    Internally working on double logarithmic E and eff scale

    """

    def __init__(self, pars=list(), degree=4, norm=False):

        self.name = "Polynom"
        self.id = self.name + "_" + hex(id(self))
        self._degree = degree

        # Building root function string
        TFString = "[0] * ([1]"

        for i in range(2, degree + 2):
            TFString += " + [%d]*x^%d" % (i, i - 1)

        TFString += ")"

        self.TF1 = TF1(self.id, TFString, 0, 0)
        _Efficiency.__init__(self, num_pars=degree + 1, pars=pars, norm=norm)

        # Builing derivatives dEff/dParameter[x]

        # See
        # http://stackoverflow.com/questions/2295290/what-do-lambda-function-closures-capture-in-python#2295372
        # http://code.activestate.com/recipes/502271/
        # for this strange constructor
        def dEff_dP(i):
            return lambda logE, fPars: self.norm * pow(logE, i)

        for i in range(0, degree + 1):
            self._dEff_dP[i] = dEff_dP(i)

    def _set_fitInput(self, fitPairs):

        ln_fitPairs = Pairs(conv_func=log)

        for p in fitPairs:
            ln_fitPairs.add(p[0], p[1])

        _Efficiency._set_fitInput(self, ln_fitPairs)

    def _get_fitInput(self):

        fitPairs = Pairs(conv_func=exp)

        for p in self._fitInput:
            fitPairs.add(p[0], p[1])

        return fitPairs

    fitInput = property(_get_fitInput, _set_fitInput)

    def normalize(self):
        # Normalize the efficiency function

        try:
            self.norm = 1.0 / exp(
                self.TF1.GetMaximum(
                    min([p[0] for p in self._fitInput]),
                    max([p[0] for p in self._fitInput]),
                )
            )
        except ZeroDivisionError:
            self.norm = 1.0

        self.TF1.SetParameter(0, self.norm)
        normfunc = TF2("norm_" + hex(id(self)), "[0]*y")
        normfunc.SetParameter(0, self.norm)
        self.TGraph.Apply(normfunc)

    def value(self, E):
        try:
            value = E.value
        except AttributeError:
            value = E

        return self.norm * exp(self.TF1.Eval(log(value), 0.0, 0.0, 0.0))

    def error(self, E):

        # TODO: this need checking
        try:
            value = E.value
        except AttributeError:
            value = E

        ln_err = _Efficiency.error(self, log(value))
        ln_eff = self.TF1.Eval(log(value), 0.0, 0.0, 0.0)
        tmp1 = self.norm * exp(ln_eff + ln_err)
        tmp2 = self.norm * exp(ln_eff - ln_err)

        error = abs(tmp1 - tmp2) / 2.0

        return self.norm * error



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
For each model implemented on the C++ side, we have a corresponding Python
class to handle fitter setup and data transfer to the Python side
"""

import ROOT
import hdtv.rootext.fit

# Base class for all peak models
class PeakModel(object):
    """
    A peak model is a function used to fit peaks. The user can choose how to fit
    its parameters (and whether to include them at all, i.e. for tails). After
    everything has been configured, the peak model produces a C++ fitter object.
    """

    def __init__(self):
        self.fGlobalParams = dict()

    def ResetGlobalParams(self):
        self.fGlobalParams.clear()

    def OptionsStr(self):
        """
        Returns a string describing the currently set parameters of the model
        """
        statstr = ""

        for name, status in self.fParStatus.items():
            # Short format for multiple values...
            if isinstance(status, list):
                statstr += "%s: " % name
                sep = ""
                for stat in status:
                    statstr += sep
                    if stat in ("free", "equal", "hold", "none", "calculated"):
                        statstr += stat
                    else:
                        statstr += "%.3f" % stat
                    sep = ", "
                statstr += "\n"

            # ... long format for a single value
            else:
                if status == "free":
                    statstr += "%s: (individually) free\n" % name
                elif status == "equal":
                    statstr += "%s: free and equal\n" % name
                elif status == "hold":
                    statstr += "%s: held at default value\n" % name
                elif status == "none":
                    statstr += "%s: none (disabled)\n" % name
                elif status == "calculated":
                    statstr += "%s: calculated from fit result\n" % name
                else:
                    statstr += "%s: fixed at %.3f\n" % (name, status)

        for name, status in self.fOptStatus.items():
            statstr += f"{name}: {status}\n"

        return statstr

    def CheckParStatusLen(self, minlen):
        """
        Checks if each parameter status provided on a peak-by-peak basis
        has at least minlen entires. Raises a ValueError with an appropriate
        error message if the check fails.
        """
        for (parname, status) in self.fParStatus.items():
            if isinstance(status, list) and len(status) < minlen:
                raise ValueError("Not enough values for status of %s" % parname)

    def ParseParamStatus(self, parname, status):
        """
        Parse a parameter status specification string
        Raises ValueError when the status is not legal for this parameter
        """
        # Case-insensitive matching
        status = status.strip().lower()

        # Check to see if status corresponds, possibly abbreviated,
        #  to a number of special keywords
        stat = None
        if len(status) < 1:
            raise ValueError

        if parname in self.fValidParStatus:
            valid_status = self.fValidParStatus
        elif parname in self.fValidOptStatus:
            valid_status = self.fValidOptStatus
        valid_strings = [p for p in valid_status[parname] if isinstance(p, (bool, str))]

        for valid_string in valid_strings:
            if str(valid_string).lower()[0 : len(status)] == status:
                stat = valid_string

        # If status was a keyword, see if this setting is legal for the
        # parameter
        if stat is not None:
            if stat not in valid_status[parname]:
                msg = "Status %s not allowed for parameter %s in peak model %s" % (
                    stat,
                    parname,
                    self.name,
                )
                raise ValueError(msg)
            return stat

        # Check if a numeric value is legal for the parameter
        if float not in valid_status[parname]:
            msg = "Invalid status %s for parameter %s in peak model %s" % (
                status,
                parname,
                self.name,
            )
            raise ValueError(msg)

        try:
            val = float(status)
        except ValueError:
            msg = "Failed to parse status specifier `%s'" % status
            raise ValueError(msg)
        return val

    def SetParameter(self, parname, status):
        """
        Set status for a certain parameter. Status is a string describing the
        desired status. Raises ValueError in case of invalid input.

        status may be single string which will be taken for all peaks, or
        list, where each item will be assigned to corresponing peak
        """
        parname = parname.strip().lower()

        try:
            if isinstance(status, list):
                if parname in self.fParStatus:
                    self.fParStatus[parname] = [
                        self.ParseParamStatus(parname, s) for s in status
                    ]
                else:
                    self.fOptStatus[parname] = [
                        self.ParseParamStatus(parname, s) for s in status
                    ]
            else:
                if parname in self.fParStatus:
                    self.fParStatus[parname] = self.ParseParamStatus(parname, status)
                else:
                    self.fOptStatus[parname] = self.ParseParamStatus(parname, status)
        except KeyError:
            raise ValueError(
                "Invalid parameter name %s for peak model %s" % (parname, self.name)
            )

    def GetParam(self, name, peak_id, pos_uncal, cal, ival=None):
        """
        Return an appropriate HDTV.Fit.Param object for the specified parameter
        """
        # See if the parameter status has been specified for each peak
        # individually, or for all peaks at once
        if isinstance(self.fParStatus[name], list):
            parStatus = self.fParStatus[name][peak_id]
        else:
            parStatus = self.fParStatus[name]
        # Switch according to parameter status
        if parStatus == "equal":
            if name not in self.fGlobalParams:
                if ival is None:
                    self.fGlobalParams[name] = self.fFitter.AllocParam()
                else:
                    self.fGlobalParams[name] = self.fFitter.AllocParam(ival)
            return self.fGlobalParams[name]
        elif parStatus == "free":
            if ival is None:
                return self.fFitter.AllocParam()
            else:
                return self.fFitter.AllocParam(ival)
        elif parStatus == "hold":
            if ival is None:
                return ROOT.HDTV.Fit.Param.Fixed()
            else:
                return ROOT.HDTV.Fit.Param.Fixed(ival)
        elif parStatus == "none":
            return ROOT.HDTV.Fit.Param.Empty()
        elif isinstance(parStatus, float):
            return ROOT.HDTV.Fit.Param.Fixed(
                self.Uncal(name, parStatus, pos_uncal, cal)
            )
        else:
            raise RuntimeError("Invalid parameter status")

    def GetOption(self, name):
        """
        Return an appropriate HDTV.Fit.Option object for the specified parameter
        """
        status = self.fOptStatus[name]
        return ROOT.HDTV.Fit.Option(type(status))(status)



#!/usr/bin/env python

from __future__ import print_function

import hdtv.efficiency as eff

e = eff.WiedenhoeverEff()

e.load("eff_wiedenhoever.par", "eff_wiedenhoever.cov")

for i in range(1, 10000, 100):
    print("Wiedenhoefer-Efficiency @", i, " keV = ", e(i))



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import re
import os
import sys
import copy

import pytest

from tests.helpers.utils import redirect_stdout, hdtvcmd

from hdtv.util import monkey_patch_ui

monkey_patch_ui()

import hdtv.cmdline
import hdtv.options
import hdtv.session

import __main__

try:
    __main__.spectra = hdtv.session.Session()
except RuntimeError:
    pass

from hdtv.plugins.specInterface import spec_interface
from hdtv.plugins.fitInterface import fit_interface
import hdtv.plugins.peakfinder

spectra = __main__.spectra

testspectrum = os.path.join(os.path.curdir, "tests", "share", "osiris_bg.spc")


@pytest.fixture(autouse=True)
def prepare():
    fit_interface.ResetFitterParameters()
    hdtv.options.Set("table", "classic")
    hdtv.options.Set("uncertainties", "short")
    spectra.Clear()
    yield
    spectra.Clear()
    fit_interface.ResetFitterParameters()


def test_cmd_fit_various():
    hdtvcmd("fit function peak activate theuerkauf")
    spec_interface.LoadSpectra(testspectrum)
    assert len(spec_interface.spectra.dict) == 1
    f, ferr = setup_fit()
    assert f == ""
    assert ferr == ""
    f, ferr = hdtvcmd("fit execute")
    assert "WorkFit on spectrum: 0" in f
    assert ".0 |" in f
    assert ".1 |" in f
    assert "2 peaks in WorkFit" in f
    f, ferr = hdtvcmd("fit store")
    assert f == "Storing workFit with ID 0"
    assert ferr == ""
    f, ferr = hdtvcmd("fit clear", "fit list")
    assert ferr == ""
    assert "Fits in Spectrum 0" in f
    f, ferr = hdtvcmd(
        "fit hide 0",
        "fit show 0",
        "fit show decomposition 0",
        "fit hide decomposition 0",
    )
    assert ferr == ""
    assert f == ""
    f, ferr = hdtvcmd("fit activate 0")
    assert "Activating fit 0" in f
    assert ferr == ""
    f, ferr = hdtvcmd("fit delete 0", "fit function peak activate ee")
    assert ferr == ""
    assert f == ""


def test_cmd_fit_peakfind():
    spec_interface.LoadSpectra(testspectrum)
    assert len(spec_interface.spectra.dict) == 1
    f, ferr = hdtvcmd("fit peakfind -a -t 0.002")
    assert "Search Peaks in region" in f
    assert "Found 68 peaks" in f
    # This was not needed before commits around ~b41833c9c66f9ba5dbdcfc6fc4468b242360641f
    # However, some small change has happened that prevents a correct fit of a single
    # double peak at ~1540 keV. When I fit it manually, using a similar fit region
    # and internal background only, it works.
    assert "WARNING: Adding invalid fit" in ferr


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@pytest.mark.parametrize("peak", ["theuerkauf", "ee"])
@pytest.mark.parametrize("bg", ["polynomial", "exponential", "interpolation"])
@pytest.mark.parametrize("integrate", ["True", "False"])
@pytest.mark.parametrize("likelihood", ["normal", "poisson"])
def test_cmd_fit_parameter(peak, bg, integrate, likelihood):
    spec_interface.LoadSpectra(testspectrum)
    f, ferr = setup_fit()
    hdtvcmd("fit marker background set 415", "fit marker background set 450")
    f, ferr = hdtvcmd(
        f"fit function peak activate {peak}",
        f"fit function background activate {bg}",
        f"fit parameter integrate {integrate}",
        f"fit parameter likelihood {likelihood}",
    )
    assert f == ""
    assert ferr == ""
    f, ferr = hdtvcmd("fit execute")
    assert "WorkFit on spectrum: 0" in f
    assert ".0 |" in f
    assert ".1 |" in f
    assert "2 peaks in WorkFit" in f

    f, ferr = hdtvcmd("fit parameter status")
    assert re.search(f"Peak model.*{peak}", f, re.M)
    assert re.search(f"Background model.*{bg}", f, re.M)
    assert f"likelihood: {likelihood}" in f
    assert f"integrate: {integrate}" in f
    assert ferr == ""

    f, ferr = hdtvcmd("fit store")
    assert f == "Storing workFit with ID 0"
    assert ferr == ""

    f, ferr = hdtvcmd("fit clear", "fit list")
    assert ferr == ""
    assert "Fits in Spectrum 0" in f

    f, ferr = hdtvcmd("fit parameter reset")
    assert ferr == ""


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_cmd_fit_copy():
    spec_interface.LoadSpectra(testspectrum)
    f, ferr = setup_fit()
    hdtvcmd("fit marker background set 415", "fit marker background set 450")
    f, ferr = hdtvcmd(
        "fit function peak activate theuerkauf",
        "fit function background activate polynomial",
        "fit parameter integrate True",
        "fit parameter likelihood poisson",
        "fit execute",
    )
    workFit = spec_interface.spectra.workFit
    newFit = copy.copy(workFit)
    assert workFit == newFit
    assert workFit.fitter == newFit.fitter


def test_interpolation_incomplete():
    spec_interface.LoadSpectra(testspectrum)
    assert len(spec_interface.spectra.dict) == 1
    f, ferr = setup_interpolation_incomplete()
    assert f == ""
    assert "Background fit failed" in ferr


def setup_interpolation_incomplete():
    return hdtvcmd(
        "fit function background activate interpolation",
        "fit marker background set 520",
        "fit marker background set 550",
        "fit marker background set 620",
        "fit marker background set 650",
        "fit execute",
    )


def setup_fit():
    return hdtvcmd(
        "fit function background activate polynomial",
        "fit parameter background 2",
        "fit marker peak set 580",
        "fit marker peak set 610",
        "fit marker background set 520",
        "fit marker background set 550",
        "fit marker background set 620",
        "fit marker background set 650",
        "fit marker region set 570",
        "fit marker region set 615",
    )


# More tests are still needed for:
#
# fit focus
# fit getlists
# fit read
# fit savelists
# fit tex
# fit write



from .theuerkaufPeak import PeakModelTheuerkauf
from .eePeak import PeakModelEE

# dictionary of available peak models
PeakModels = dict()
PeakModels["theuerkauf"] = PeakModelTheuerkauf
PeakModels["ee"] = PeakModelEE



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import pytest

import hdtv.util


@pytest.mark.parametrize(
    "cmdline, segs, last_suffix",
    [
        ("", [], ""),
        ("a", [["a"]], ""),
        ("a;", [["a"]], ""),
        ("a;b", [["a"], ["b"]], ""),
        ("a ;b", [["a"], ["b"]], ""),
        ("a; b", [["a"], ["b"]], ""),
        ("a ; b", [["a"], ["b"]], ""),
        (" a ; b", [["a"], ["b"]], ""),
        ("a;b c", [["a"], ["b", "c"]], ""),
        ('a;b "c d"', [["a"], ["b", "c d"]], ""),
        ('a "b', [["a", "b"]], '"'),
        ('a ";', [["a", ";"]], '"'),
        ("a#b", [["a"]], ""),
        ('a "" a', [["a", "", "a"]], ""),
    ],
)
def test_SplitCmdlines(cmdline, segs, last_suffix):
    res_segs, res_last_suffix = hdtv.util.SplitCmdlines(cmdline)
    assert segs == res_segs
    assert last_suffix == res_last_suffix



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import os

from scipy.interpolate import InterpolatedUnivariateSpline
import numpy as np

import ROOT
import hdtv.color
import hdtv.rootext.mfile
import hdtv.rootext.calibration
import hdtv.rootext.display
import hdtv.rootext.fit

from hdtv.drawable import Drawable
from hdtv.specreader import SpecReader, SpecReaderError
from hdtv.cal import CalibrationFitter
from hdtv.util import LockViewport

# Don't add created spectra to the ROOT directory
ROOT.TH1.AddDirectory(ROOT.kFALSE)


def HasPrimitiveBinning(hist):
    if hist.GetNbinsX() != (hist.GetXaxis().GetXmax() - hist.GetXaxis().GetXmin()):
        return False
    for bin in range(0, hist.GetNbinsX()):
        if hist.GetBinWidth(bin) != 1.0:
            return False
    return True


class Histogram(Drawable):
    """
    Histogram object

    This class is hdtvs wrapper around a ROOT histogram. It adds a calibration,
    plus some internal management for drawing the histogram to the hdtv spectrum
    viewer.
    """

    def __init__(self, hist, color=hdtv.color.default, cal=None):
        Drawable.__init__(self, color, cal)
        self._norm = 1.0
        self._ID = None
        self.effCal = None
        self.typeStr = "spectrum"
        self.cal = cal

        if cal is None:
            self.SetHistWithPrimitiveBinning(hist)
        else:
            self._hist = hist

    def __str__(self):
        return self.name

    def __copy__(self):
        # call C++ copy constructor
        hist = self._hist.__class__(self._hist)
        # create new spectrum object
        return Histogram(hist, color=self.color, cal=self.cal)

    # hist property
    def _set_hist(self, hist):
        self._hist = hist
        if self.displayObj:
            self.displayObj.SetHist(self._hist)

    def _get_hist(self):
        return self._hist

    hist = property(_get_hist, _set_hist)

    # name property
    def _get_name(self):
        if self._hist:
            return self._hist.GetName()

    def _set_name(self, name):
        self._hist.SetName(name)

    name = property(_get_name, _set_name)

    # norm property
    def _set_norm(self, norm):
        self._norm = norm
        if self.displayObj:
            self.displayObj.SetNorm(norm)

    def _get_norm(self):
        return self._norm

    norm = property(_get_norm, _set_norm)

    @property
    def info(self):
        """
        Return a string describing this spectrum
        """
        s = "Spectrum type: %s\n" % self.typeStr
        if not self._hist:
            return s
        s += "Name: %s\n" % str(self)
        s += "Nbins: %d\n" % self._hist.GetNbinsX()
        xmin = self._hist.GetXaxis().GetXmin()
        xmax = self._hist.GetXaxis().GetXmax()
        if self.cal and not self.cal.IsTrivial():
            s += "Xmin: %.2f (cal)  %.2f (uncal)\n" % (self.cal.Ch2E(xmin), xmin)
            s += "Xmax: %.2f (cal)  %.2f (uncal)\n" % (self.cal.Ch2E(xmax), xmax)
        else:
            s += "Xmin: %.2f\n" % xmin
            s += "Xmax: %.2f\n" % xmax

        if not self.cal or self.cal.IsTrivial():
            s += "Calibration: none\n"
        elif isinstance(self.cal, ROOT.HDTV.Calibration):
            s += "Calibration: Polynomial, degree %d\n" % self.cal.GetDegree()
        else:
            s += "Calibration: unknown\n"
        return s

    # TODO: sumw2 function should be called at some point for correct error
    # handling
    def Plus(self, spec):
        """
        Add other spectrum to this one
        """
        # If the spectra have the same calibration (~= have the same binning),
        # the root build-in add can be used
        if self.cal == spec.cal or (self.cal.IsTrivial() and spec.cal.IsTrivial()):
            hdtv.ui.info("Adding binwise")
            self._hist.Add(spec._hist, 1.0)
        # If the binning is different, determine the amount to add to each bin
        # by integrating the other spectrum
        else:
            hdtv.ui.info("Adding calibrated")
            nbins = self._hist.GetNbinsX()
            for n in range(0, nbins):
                integral = ROOT.HDTV.TH1IntegrateWithPartialBins(
                    spec._hist,
                    spec.cal.E2Ch(self.cal.Ch2E(n - 0.5)),
                    spec.cal.E2Ch(self.cal.Ch2E(n + 0.5)),
                )
                # Note: Can't use Fill due to bin errors?
                self._hist.SetBinContent(
                    n + 1, self._hist.GetBinContent(n + 1) + integral
                )

        # update display
        if self.displayObj:
            self.displayObj.SetHist(self._hist)
        self.typeStr = "spectrum, modified (sum)"

    def Minus(self, spec):
        """
        Substract other spectrum from this one
        """
        # If the spectra have the same calibration (~= have the same binning),
        # the root build-in add can be used
        if self.cal == spec.cal or (self.cal.IsTrivial() and spec.cal.IsTrivial()):
            hdtv.ui.info("Adding binwise")
            self._hist.Add(spec._hist, -1.0)
        # If the binning is different, determine the amount to add to each bin
        # by integrating the other spectrum
        else:
            hdtv.ui.info("Adding calibrated")
            nbins = self._hist.GetNbinsX()
            for n in range(0, nbins):
                integral = ROOT.HDTV.TH1IntegrateWithPartialBins(
                    spec._hist,
                    spec.cal.E2Ch(self.cal.Ch2E(n - 0.5)),
                    spec.cal.E2Ch(self.cal.Ch2E(n + 0.5)),
                )
                # Note: Can't use Fill due to bin errors?
                self._hist.SetBinContent(
                    n + 1, self._hist.GetBinContent(n + 1) - integral
                )

        # update display
        if self.displayObj:
            self.displayObj.SetHist(self._hist)
        self.typeStr = "spectrum, modified (difference)"

    def Multiply(self, factor):
        """
        Multiply spectrum with factor
        """
        self._hist.Scale(factor)
        # update display
        if self.displayObj:
            self.displayObj.SetHist(self._hist)
        self.typeStr = "spectrum, modified (multiplied)"

    def Rebin(self, ngroup, calibrate=True):
        """
        Rebin spectrum by adding ngroup bins into one
        """
        bins = self._hist.GetNbinsX()
        self._hist.RebinX(ngroup)
        self._hist.GetXaxis().SetLimits(0, bins / ngroup)
        # update display
        if self.displayObj:
            self.displayObj.SetHist(self._hist)
        # update calibration
        if calibrate:
            if not self.cal:
                self.cal.SetCal(0.0, 1.0)
            self.cal.Rebin(ngroup)
            self.displayObj.SetCal(self.cal)
            hdtv.ui.info("Calibration updated for rebinned spectrum")
        self.typeStr = f"spectrum, modified (rebinned, ngroup={ngroup})"

    def Calbin(
        self, binsize: float = 1.0, spline_order: int = 3, use_tv_binning: bool = True
    ):
        """
        Rebin spectrum to match calibration unit

        Args:
            binsize: Size of calibrated bins
            spline_order: Order of the spline interpolation (default: 3)
            use_tv_binning: Center first bin on 0. (True) or
                lower edge of first bin on 0. (False).
        """

        nbins_old = self._hist.GetNbinsX()
        lower_old = self.cal.Ch2E(0)
        upper_old = self.cal.Ch2E(nbins_old - 1)

        nbins = int(np.ceil(upper_old / binsize)) + 1
        if use_tv_binning:
            lower = -0.5 * binsize
            upper = 0.5 * binsize + (upper_old // nbins) * (nbins - 1)
        else:
            lower = 0.0
            upper = binsize + (upper_old // nbins) * (nbins - 1)

        # Create new histogram with number of bins equal
        # to the calibrated range of the old histogram
        # Always -0.5 to create standard tv-type histogram
        newhist = ROOT.TH1D(
            self._hist.GetName(), self._hist.GetTitle(), nbins, -0.5, nbins - 0.5
        )

        input_bins_center, input_hist = np.transpose(
            [
                [
                    self.cal.Ch2E(n - 1),
                    self._hist.GetBinContent(n)
                    / (self.cal.Ch2E(n) - self.cal.Ch2E(n - 1)),
                ]
                for n in range(1, self._hist.GetNbinsX() + 1)
            ]
        )

        output_bins_low = np.arange(nbins) * binsize + lower
        output_bins_high = output_bins_low + binsize

        inter = InterpolatedUnivariateSpline(
            input_bins_center, input_hist, k=spline_order
        )

        inter_integral_v = np.vectorize(inter.integral)
        output_hist = np.maximum(
            inter_integral_v(output_bins_low, output_bins_high), 0.0
        )

        # Suppress bins outside of original histogram range
        min_bin = int((lower_old - lower) / binsize)
        output_hist[:min_bin] = np.zeros(min_bin)

        for i in range(0, nbins):
            newhist.SetBinContent(i + 1, output_hist[i])

        self._hist = newhist
        if use_tv_binning:
            if binsize != 1.0 or self.cal:
                self.cal.SetCal(0, binsize)
        else:
            self.cal.SetCal(binsize / 2, binsize)
        # update display
        if self.displayObj:
            self.displayObj.SetHist(self._hist)
        # update calibration
        self.displayObj.SetCal(self.cal)
        hdtv.ui.info(f"Rebinned to calibration unit (binsize={binsize}).")

    def Poisson(self):
        """
        Randomize each bin content assuming a Poissonian distribution.
        """
        for i in range(0, self._hist.GetNbinsX() + 1):
            counts = self._hist.GetBinContent(i)
            # error = self._hist.GetBinError(i)
            varied = np.random.poisson(counts)
            self._hist.SetBinContent(i, varied)
        if self.displayObj:
            self.displayObj.SetHist(self._hist)

    def Draw(self, viewport):
        """
        Draw this spectrum to the viewport
        """

        if self.viewport is not None and not self.viewport == viewport:
            # Unlike the DisplaySpec object of the underlying implementation,
            # Spectrum() objects can only be drawn on a single viewport
            raise RuntimeError("Spectrum can only be drawn on a single viewport")
        self.viewport = viewport
        # Lock updates
        with LockViewport(self.viewport):
            # Show spectrum
            if self.displayObj is None and self._hist is not None:
                if self.active:
                    color = self._activeColor
                else:
                    color = self._passiveColor
                self.displayObj = ROOT.HDTV.Display.DisplaySpec(self._hist, color)
                self.displayObj.SetNorm(self.norm)
                self.displayObj.Draw(self.viewport)
                # add calibration
                if self.cal:
                    self.displayObj.SetCal(self.cal)
                # and ID
                if self.ID is not None:
                    ID = str(self.ID).strip(".")
                    self.displayObj.SetID(ID)

    def WriteSpectrum(self, fname, fmt):
        """
        Write the spectrum to file
        """
        fname = os.path.expanduser(fname)
        try:
            SpecReader.WriteSpectrum(self._hist, fname, fmt)
        except SpecReaderError as msg:
            hdtv.ui.error("Failed to write spectrum: %s (file: %s)" % (msg, fname))
            return False
        return True

    def SetHistWithPrimitiveBinning(self, hist, caldegree=4, silent=False):
        log = hdtv.ui.debug if silent else hdtv.ui.info
        if HasPrimitiveBinning(hist):
            self._hist = hist
        else:
            log(
                hist.GetName()
                + " unconventional binning detected. Converting and trying to create calibration using a polynomial of order "
                + str(caldegree)
                + " ..."
            )
            self._hist = ROOT.TH1D(
                hist.GetName(), hist.GetTitle(), hist.GetNbinsX(), 0, hist.GetNbinsX()
            )
            if caldegree:
                cf = CalibrationFitter()
            # TODO: Slow
            for bin in range(0, hist.GetNbinsX()):
                if caldegree:
                    cf.AddPair(bin, hist.GetXaxis().GetBinUpEdge(bin))
                self._hist.SetBinContent(bin, hist.GetBinContent(bin))
                # Original comment by JM in commit #dd438b7c44265072bf8b0528170cecc95780e38c:
                # "TODO: Copy Errors?"
                #
                # Edit by UG: It makes sense to simply copy the uncertainties. There are two
                # possible cases:
                # 1. The ROOT histogram contains user-defined uncertainties per bin that can
                #    be retrieved by calling hist.GetBinError(). In this case, it can be
                #    assumed that the user knew what he was doing when the uncertainties
                #    were assigned.
                # 2. The ROOT histogram contains no user-defined uncertainties. In this case,
                #    a call of hist.GetBinError() will return the square root of the bin
                #    content, which is a sensible assumption.
                #
                # Since text spectra are loaded in a completely analogous way, implicitly
                # assuming that the uncertainties are Poissonian, there is no need to issue
                # an additional warning.
                self._hist.SetBinError(bin, hist.GetBinError(bin))
            if caldegree:
                cf.FitCal(caldegree)
                self.cal = cf.calib


class FileHistogram(Histogram):
    """
    File spectrum object

    A spectrum that comes from a file in any of the formats supported by hdtv.
    """

    def __init__(self, fname, fmt=None, color=hdtv.color.default, cal=None):
        """
        Read a spectrum from file
        """
        # check if file exists
        try:
            os.path.exists(fname)
        except OSError:
            hdtv.ui.error("File %s not found" % fname)
            raise
        # call to SpecReader to get the hist
        try:
            hist = SpecReader.GetSpectrum(fname, fmt)
        except SpecReaderError as msg:
            hdtv.ui.error(str(msg))
            raise
        self.fmt = fmt
        self.filename = fname
        Histogram.__init__(self, hist, color, cal)
        self.typeStr = "spectrum, read from file"

    @property
    def info(self):
        # get the info property of the baseclass
        s = super(FileHistogram, self).info
        s += "Filename: %s\n" % self.filename
        if self.fmt:
            s += "File format: %s\n" % self.fmt
        else:
            s += "File format: autodetected\n"
        return s

    def Refresh(self):
        """
        Reload the spectrum from disk
        """
        try:
            os.path.exists(self.filename)
        except OSError:
            hdtv.ui.warning("File %s not found, keeping previous data" % self.filename)
            return
        # call to SpecReader to get the hist
        try:
            hist = SpecReader.GetSpectrum(self.filename, self.fmt)
        except SpecReaderError as msg:
            hdtv.ui.warning(
                "Failed to load spectrum: %s (file: %s), keeping previous data"
                % (msg, self.filename)
            )
            return
        self.hist = hist


class CutHistogram(Histogram):
    def __init__(self, hist, axis, gates, color=hdtv.color.default, cal=None):
        Histogram.__init__(self, hist, color, cal)
        self.gates = gates
        self.axis = axis

    @property
    def info(self):
        s = super(CutHistogram, self).info
        s += "cut "
        s += "on %s axis gate: " % self.axis
        for i in range(len(self.gates)):
            g = self.gates[i]
            s += "%d - %d " % (g.p1.pos_cal, g.p2.pos_cal)
            if not i == len(self.gates):
                "and"
        return s


class THnSparseWrapper(object):
    """
    Wrapper around a 2d THnSparse object, providing ProjectionX and
    ProjectionY.
    """

    def __init__(self, hist):
        if not (isinstance(hist, ROOT.THnSparse) and hist.GetNdimensions() == 2):
            raise RuntimeError("Class needs a THnSparse histogram of dimension 2")
        self.__dict__["_hist"] = hist

    def __setattr__(self, name, value):
        self.__dict__["_hist"].__setattr__(name, value)

    def __getattr__(self, name):
        return getattr(self.__dict__["_hist"], name)

    def GetXaxis(self):
        return self._hist.GetAxis(0)

    def GetYaxis(self):
        return self._hist.GetAxis(1)

    def ProjectionX(self, name, b1, b2, opt):
        a = self._hist.GetAxis(1)
        if b1 > b2:
            a.SetRange(0, a.GetNbins())
        else:
            a.SetRange(b1, b2)

        proj = self._hist.Projection(0, opt)
        a.SetRange(0, a.GetNbins())

        proj.SetName(name)
        return proj

    def ProjectionY(self, name, b1, b2, opt):
        a = self._hist.GetAxis(0)
        if b1 > b2:
            a.SetRange(0, a.GetNbins())
        else:
            a.SetRange(b1, b2)

        proj = self._hist.Projection(1, opt)
        a.SetRange(0, a.GetNbins())

        proj.SetName(name)
        return proj


class Histo2D(object):
    def __init__(self):
        pass

    @property
    def name(self):
        return "generic 2D histogram"

    @property
    def xproj(self):
        return None

    @property
    def yproj(self):
        return None

    def ExecuteCut(self, regionMarkers, bgMarkers, axis):
        return None


class RHisto2D(Histo2D):
    """
    ROOT TH2-backed matrix for projection
    """

    def __init__(self, rhist):
        self.rhist = rhist

        # Lazy generation of projections
        self._prx = None
        self._pry = None

    @property
    def name(self):
        return self.rhist.GetName()

    @property
    def xproj(self):
        if self._prx is None:
            name = self.rhist.GetName() + "_prx"
            self._prx = self.rhist.ProjectionX(name, 0, -1, "e")
            # do not store the Histogram object here because of garbage
            # collection
            prx = Histogram(self._prx)
            prx.typeStr = "x projection"
        return prx

    @property
    def yproj(self):
        if self._pry is None:
            name = self.rhist.GetName() + "_pry"
            self._pry = self.rhist.ProjectionY(name, 0, -1, "e")
            # do not store the Histogram object here because of garbage
            # collection
            pry = Histogram(self._pry)
            pry.typeStr = "y projection"
        return pry

    def ExecuteCut(self, regionMarkers, bgMarkers, axis):
        # _axis_ is the axis the markers refer to, so we project on the *other*
        # axis. We call _axis_ the cut axis and the other axis the projection
        # axis. If the matrix is symmetric, this does not matter, so _axis_ is
        # "0" and the implementation can choose.

        if len(regionMarkers) < 1:
            raise RuntimeError("Need at least one gate for cut")

        if axis == "0":
            axis = "x"

        if axis not in ("x", "y"):
            raise ValueError("Bad value for axis parameter")

        if axis == "x":
            cutAxis = self.rhist.GetXaxis()
            projector = self.rhist.ProjectionY
        else:
            cutAxis = self.rhist.GetYaxis()
            projector = self.rhist.ProjectionX

        b1 = cutAxis.FindBin(regionMarkers[0].p1.pos_uncal)
        b2 = cutAxis.FindBin(regionMarkers[0].p2.pos_uncal)

        name = self.rhist.GetName() + "_cut"
        rhist = projector(name, min(b1, b2), max(b1, b2), "e")
        # Ensure proper garbage collection for ROOT histogram objects
        ROOT.SetOwnership(rhist, True)

        numFgBins = abs(b2 - b1) + 1
        for r in regionMarkers[1:]:
            b1 = cutAxis.FindBin(r.p1.pos_uncal)
            b2 = cutAxis.FindBin(r.p2.pos_uncal)
            numFgBins += abs(b2 - b1) + 1

            tmp = projector("proj_tmp", min(b1, b2), max(b1, b2), "e")
            ROOT.SetOwnership(tmp, True)
            rhist.Add(tmp, 1.0)

        bgBins = []
        numBgBins = 0
        for b in bgMarkers:
            b1 = cutAxis.FindBin(b.p1.pos_uncal)
            b2 = cutAxis.FindBin(b.p2.pos_uncal)
            numBgBins += abs(b2 - b1) + 1
            bgBins.append((min(b1, b2), max(b1, b2)))

        if numBgBins > 0:
            bgFactor = -float(numFgBins) / float(numBgBins)

            for b in bgBins:
                tmp = projector("proj_tmp", b[0], b[1], "e")
                ROOT.SetOwnership(tmp, True)
                rhist.Add(tmp, bgFactor)

        hist = CutHistogram(rhist, axis, regionMarkers)
        hist.typeStr = "cut"
        return hist


class MHisto2D(Histo2D):
    """
    MFile-backed matrix for projection
    """

    def __init__(self, fname, sym):
        # check if file exists
        try:
            os.stat(fname)
        except OSError as error:
            hdtv.ui.error(str(error))
            raise

        self.GenerateFiles(fname, sym)

        basename = self.GetBasename(fname)

        # call to SpecReader to get the hist
        try:
            self.vmatrix = SpecReader.GetVMatrix(fname)
        except SpecReaderError as msg:
            hdtv.ui.error(str(msg))
            raise

        self._xproj = FileHistogram(basename + ".prx")
        self._xproj.typeStr = "Projection"

        if sym:
            self._yproj = None
            self.tvmatrix = self.vmatrix  # Fixme
        else:
            self._yproj = FileHistogram(basename + ".pry")
            self._yproj.typeStr = "Projection"

            try:
                self.tvmatrix = SpecReader.GetVMatrix(basename + ".tmtx")
            except SpecReaderError as msg:
                hdtv.ui.error(str(msg))
                raise

        self.filename = fname

    @property
    def xproj(self):
        return self._xproj

    @property
    def yproj(self):
        return self._yproj

    def ExecuteCut(self, regionMarkers, bgMarkers, axis):
        # _axis_ is the axis the markers refer to, so we project on the *other*
        # axis. We call _axis_ the cut axis and the other axis the projection
        # axis. If the matrix is symmetric, this does not matter, so _axis_ is
        # "0" and the implementation can choose.

        if len(regionMarkers) < 1:
            raise RuntimeError("Need at least one gate for cut")

        if axis == "0":
            axis = "x"

        if axis not in ("x", "y"):
            raise ValueError("Bad value for axis parameter")

        if axis == "x":
            # FIXME: Calibrations for gated spectra asym/sym
            thiscal = self._xproj.cal
            if self._yproj:
                othercal = self._yproj.cal
            else:
                othercal = self._xproj.cal
            matrix = self.tvmatrix
        else:
            thiscal = self._yproj.cal
            othercal = self._xproj.cal
            matrix = self.vmatrix

        matrix.ResetRegions()

        for r in regionMarkers:
            # FIXME: The region markers are not used correctly in many parts
            # of the code. Workaround by explicitly using the cal here
            b1 = matrix.FindCutBin(thiscal.E2Ch(r.p1.pos_cal))
            b2 = matrix.FindCutBin(thiscal.E2Ch(r.p2.pos_cal))
            matrix.AddCutRegion(b1, b2)

        for b in bgMarkers:
            b1 = matrix.FindCutBin(thiscal.E2Ch(b.p1.pos_cal))
            b2 = matrix.FindCutBin(thiscal.E2Ch(b.p2.pos_cal))
            matrix.AddBgRegion(b1, b2)

        name = self.filename + "_cut"
        rhist = matrix.Cut(name, name)
        # Ensure proper garbage collection for ROOT histogram objects
        ROOT.SetOwnership(rhist, True)

        hist = CutHistogram(rhist, axis, regionMarkers)
        hist.typeStr = "cut"
        hist._cal = othercal
        return hist

    def GetBasename(self, fname):
        if fname.endswith(".mtx") or fname.endswith(".mtx"):
            return fname[:-4]
        else:
            return fname

    def GenerateFiles(self, fname, sym):
        """
        Generate projection(s) and possibly transpose (for asymmetric matrices),
        if they do not exist yet.
        """
        basename = self.GetBasename(fname)

        # Generate projection(s)
        prx_fname = basename + ".prx"
        pry_fname = ""
        if os.path.exists(prx_fname):
            hdtv.ui.info("Using %s for x projection" % prx_fname)
            prx_fname = ""

        if not sym:
            pry_fname = basename + ".pry"
            if os.path.exists(pry_fname):
                hdtv.ui.info("Using %s for y projection" % pry_fname)
                pry_fname = ""

        if prx_fname or pry_fname:
            errno = ROOT.MatOp.Project(fname, prx_fname, pry_fname)
            if errno != ROOT.MatOp.ERR_SUCCESS:
                raise RuntimeError("Project: " + ROOT.MatOp.GetErrorString(errno))

            if prx_fname:
                hdtv.ui.info("Generated x projection: %s" % prx_fname)
            if pry_fname:
                hdtv.ui.info("Generated y projection: %s" % pry_fname)

        # Generate transpose
        if not sym:
            trans_fname = basename + ".tmtx"
            if os.path.exists(trans_fname):
                hdtv.ui.info("Using %s for transpose" % trans_fname)
            else:
                errno = ROOT.MatOp.Transpose(fname, trans_fname)
                if errno != ROOT.MatOp.ERR_SUCCESS:
                    raise RuntimeError("Transpose: " + ROOT.MatOp.GetErrorString(errno))
                hdtv.ui.info("Generated transpose: %s" % trans_fname)
